'''
A work in progress to abstract over the messy shared dictionary scheme used in
draw_tree.DrawTreeNode
'''

import weakref
from itertools import chain

import numpy as np
from matplotlib.path import Path

try:
    from matplotlib.path import get_paths_extents
except ImportError:
    # TODO: Check why this was deprecated, beyond redundancy
    def get_paths_extents(paths):
        from matplotlib.path import _path
        from matplotlib.transforms import BBox, Affine2D

        transforms = []
        if len(paths) == 0:
            raise ValueError("No paths provided")
        return Bbox.from_extents(*_path.get_path_collection_extents(
            Affine2D(), paths, transforms, [], Affine2D()))


chain_iterable = chain.from_iterable


def ensure_iterable(obj):
    try:
        return [x for x in obj if x is not None]
    except Exception:
        if obj is None:
            return []
        return [obj]


def flatten(iterable):
    return list(chain_iterable(map(ensure_iterable, iterable)))


class GraphicalPatch(object):  # pragma: no cover
    def __init__(self, shape_patches, text_patches, parent=None, **kwargs):
        self.shape_patches = shape_patches
        self.text_patches = text_patches
        self.parent = weakref.ref(parent) if parent is not None else None
        self.options = kwargs

    def transform_shapes(self, transform):
        for shape in self.shapes():
            shape.set_transform(transform)

    def transform_text(self, transform):
        for text in self.text():
            text.set_transform(transform)

    def shapes(self):
        for shape in self.shape_patches:
            if shape is not None:
                yield shape

    def text(self):
        for t in self.text_patches:
            if t is not None:
                yield t

    def set_gid(self, tree_id, node_id):
        for i, shape in enumerate(self.shapes()):
            shape.set_gid("%s-%s-%d-shape" % (tree_id, node_id, i))
        for i, text in enumerate(self.text()):
            text.set_gid("%s-%s-%d-text" % (tree_id, node_id, i))

    def centroid(self):
        point = np.zeros(2)
        c = 0
        for shape in flatten(self.shapes()):
            path_ = shape.get_path()
            for p in path_.vertices:
                point += p
                c += 1
        return point / c

    def get_paths(self):
        result = []
        for shape in flatten(self.shapes()):
            result.append(shape.get_path())
        return result

    def get_bbox(self):
        return get_paths_extents(self.get_paths())


class MonosaccharidePatch(GraphicalPatch):
    def __init__(self, saccharide_shape=(None,), substituent_text=(None,), saccharide_label=(None,),
                 parent=None, **kwargs):
        saccharide_shape = ensure_iterable(saccharide_shape)
        substituent_text = ensure_iterable(substituent_text)
        saccharide_label = ensure_iterable(saccharide_label)
        self.saccharide_shape = flatten(saccharide_shape)
        self.substituent_text = flatten(substituent_text)
        self.saccharide_label = saccharide_label
        super(MonosaccharidePatch, self).__init__(
            (saccharide_shape),
            (substituent_text + saccharide_label),
            parent, **kwargs)

    def set_gid(self, tree_id, node_id):
        node_id = str(node_id)
        for el in self.saccharide_shape:
            el.set_gid("%s-%s-node" % (tree_id, node_id))
        for i, el in enumerate(self.substituent_text):
            el.set_gid("%s-%s-subst-%d" % (tree_id, node_id, i))
        for el in self.saccharide_label:
            el.set_gid("%s-%s-label" % (tree_id, node_id))

    def add_substituents(self, substituents):
        substituents = flatten(ensure_iterable(substituents))
        self.substituent_text.extend(substituents)
        self.text_patches.extend(substituents)

    def __iter__(self):
        yield self.saccharide_shape
        yield self.substituent_text
        yield self.saccharide_label


class EdgePatch(GraphicalPatch):  # pragma: no cover
    def __init__(self, line, parent_label=None, child_label=None, parent=None, **kwargs):
        self.line_shape = ensure_iterable(line)
        self.parent_label = ensure_iterable(parent_label)
        self.child_label = ensure_iterable(child_label)
        super(EdgePatch, self).__init__(
            line, (parent_label, child_label), parent=parent, **kwargs)


class GraphicalAggregate(list):
    def __init__(self, elements):
        list.__init__(self, elements)



# Adapted from https://github.com/llimllib/pymag-trees/blob/master/buchheim.py

from .geometry import TreeLayoutBase


class BalancedTreeLayout(TreeLayoutBase):

    def layout_tree(self, **kwargs):
        buchheim(self.root)


def buchheim(tree, visited=None):
    if visited is None:
        visited = set()
    dt = first_walk(tree, visited=visited)
    visited = set()
    min = second_walk(dt, visited=visited)
    if min < 0:
        third_walk(dt, -min)
    return dt


layout = buchheim


def third_walk(tree, n, visited=None):  # pragma: no cover
    if visited is None:
        visited = set()

    tree.x += n

    if tree.id in visited:
        return

    visited.add(tree.id)

    for c in tree.children:
        third_walk(c, n, visited=visited)


def first_walk(v, distance=1., visited=None):
    if visited is None:  # pragma: no cover
        visited = set()
    if v.id in visited:  # pragma: no cover
        return v
    visited.add(v.id)
    if len(v.children) == 0:
        if v.lmost_sibling:
            v.x = v.lbrother().x + distance
        else:
            v.x = 0.
    else:
        default_ancestor = v.children[0]
        for w in v.children:
            first_walk(w, distance, visited)
            default_ancestor = apportion(w, default_ancestor, distance)
        execute_shifts(v)

        midpoint = (v.children[0].x + v.children[-1].x) / 2

        # ell = v.children[0]
        # arr = v.children[-1]
        w = v.lbrother()
        if w:
            v.x = w.x + distance
            v.mod = v.x - midpoint
        else:
            v.x = midpoint
    return v


def apportion(v, default_ancestor, distance):
    w = v.lbrother()
    if w is not None:
        # in buchheim notation:
        # i == inner; o == outer; r == right; l == left; r = +; l = -
        vir = vor = v
        vil = w
        vol = v.lmost_sibling
        sir = sor = v.mod
        sil = vil.mod
        sol = vol.mod
        while vil.right() and vir.left():
            vil = vil.right()
            vir = vir.left()
            vol = vol.left()
            vor = vor.right()
            vor.ancestor = v
            shift = (vil.x + sil) - (vir.x + sir) + distance
            if shift > 0:  # pragma: no cover
                move_subtree(ancestor(vil, v, default_ancestor), v, shift)
                sir = sir + shift
                sor = sor + shift
            sil += vil.mod
            sir += vir.mod
            sol += vol.mod
            sor += vor.mod
        if vil.right() and not vor.right():
            vor.thread = vil.right()
            vor.mod += sil - sor
        else:
            if vir.left() and not vol.left():
                vol.thread = vir.left()
                vol.mod += sir - sol
            default_ancestor = v
    return default_ancestor


def move_subtree(wl, wr, shift):  # pragma: no cover
    subtrees = wr.number - wl.number
    wr.change -= shift / subtrees
    wr.shift += shift
    wl.change += shift / subtrees
    wr.x += shift
    wr.mod += shift


def execute_shifts(v):
    shift = change = 0
    for w in v.children[::-1]:
        w.x += shift
        w.mod += shift
        change += w.change
        shift += w.shift + change


def ancestor(vil, v, default_ancestor):  # pragma: no cover
    # the relevant text is at the bottom of page 7 of
    # "Improving Walker's Algorithm to Run in Linear Time" by Buchheim et al, (2002)
    # http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.16.8757&rep=rep1&type=pdf
    if vil.ancestor in v.parent.children:
        return vil.ancestor
    else:
        return default_ancestor


def second_walk(v, m=0, depth=0, min=None, visited=None):
    if visited is None:
        visited = set()
    v.x += m
    v.y = depth

    if v.id in visited:
        return v.x

    visited.add(v.id)

    if min is None or v.x < min:
        min = v.x

    for w in v.children:
        min = second_walk(w, m + v.mod, depth + 1, min, visited=visited)

    return min



# Credit to Pyteomics - http://pythonhosted.org/pyteomics - for majority of design
import re
import math
from collections import defaultdict
from .mass_dict import nist_mass
from .base import ChemicalCompositionError, composition_factory

import os

use_cython = bool(int(os.environ.get("USE_CYTHON_COMPOSITION", 1)))

try:
    from .ccomposition import CComposition, calculate_mass as ccalculate_mass
except ImportError:  # pragma: no cover
    use_cython = False

# Forward Declaration

_isotope_string = r'^([A-Z][a-z+]*)(?:\[(\d+)\])?$'
_atom = r'([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?'
_formula = r'^({})*$'.format(_atom)
formula_pattern = re.compile(_formula)


def _make_isotope_string(element_name, isotope_num):
    """Form a string label for an isotope."""
    if isotope_num == 0:
        return element_name
    else:
        return '%s[%d]' % (element_name, isotope_num)


def _parse_isotope_string(label):
    """Parse an string with an isotope label and return the element name and
    the isotope number.

    >>> _parse_isotope_string('C')
    ('C', 0)
    >>> _parse_isotope_string('C[12]')
    ('C', 12)
    """
    element_name, num = re.match(_isotope_string, label).groups()
    isotope_num = int(num) if num else 0
    return element_name, isotope_num


class PComposition(defaultdict):
    '''Represent arbitrary elemental compositions'''
    def __str__(self):   # pragma: no cover
        return 'Composition({})'.format(dict.__repr__(self))

    def __repr__(self):  # pragma: no cover
        return str(self)

    def __iadd__(self, other):
        for elem, cnt in (other.items()):
            self[elem] += cnt
        self._mass_args = None
        return self

    def __add__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] += cnt
        return result

    def __radd__(self, other):
        return self + other

    def __isub__(self, other):
        for elem, cnt in other.items():
            self[elem] -= cnt
        self._mass_args = None
        return self

    def __sub__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] -= cnt
        return result

    def __rsub__(self, other):
        return (self - other) * (-1)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise ChemicalCompositionError(
                'Cannot multiply Composition by non-integer',
                other)

        prod = {}
        for k, v in self.items():
            prod[k] = v * other

        return Composition(prod)

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        if not isinstance(other, dict):
            return False
        self_items = set([i for i in self.items() if i[1]])
        other_items = set([i for i in other.items() if i[1]])
        return self_items == other_items

    def __neg__(self):
        return -1 * self

    def __reduce__(self):
        return composition_factory, (list(self),), self.__getstate__()

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, d):
        self._from_dict(d)

    # Override the default behavior, if a key is not present
    # do not initialize it to 0.
    def __missing__(self, key):
        return 0

    def __setitem__(self, key, value):
        if isinstance(value, float):
            value = int(round(value))
        elif not isinstance(value, (int)):
            raise ChemicalCompositionError(
                'Only integers allowed as values in Composition, got {}.'.format(type(value).__name__))
        if value:  # Will not occur on 0 as 0 is falsey AND an integer
            super(PComposition, self).__setitem__(key, value)
        elif key in self:
            del self[key]
        self._mass_args = None

    def clone(self):
        return PComposition(self)

    copy = clone

    def update(self, *args, **kwargs):
        dict.update(self, *args, **kwargs)
        self._mass_args = None

    def _from_formula(self, formula, mass_data):
        if '(' in formula:
            return self._from_formula_parens(formula, mass_data)
        if not formula_pattern.match(formula):
            raise ChemicalCompositionError('Invalid formula: ' + formula)
        for elem, isotope, number in re.findall(_atom, formula):
            if elem not in mass_data:
                raise ChemicalCompositionError('Unknown chemical element: ' + elem)
            self[_make_isotope_string(elem, int(isotope) if isotope else 0)
                 ] += int(number) if number else 1

    def _from_formula_parens(self, formula, mass_data):  # pragma: no cover
        # Parsing a formula backwards.
        prev_chem_symbol_start = len(formula)
        i = len(formula) - 1

        seek_mode = 0
        parse_stack = ""
        resolve_stack = []
        group_coef = None

        while i >= 0:
            if seek_mode < 1:
                if (formula[i] == ")"):
                    seek_mode += 1
                    if i + 1 == prev_chem_symbol_start:
                        group_coef = 1
                    elif formula[i + 1].isdigit():
                        group_coef = int(formula[i + 1:prev_chem_symbol_start])
                    i -= 1
                    continue
                # Read backwards until a non-number character is met.
                if (formula[i].isdigit() or formula[i] == '-'):
                    i -= 1
                    continue

                else:
                    # If the number of atoms is omitted then it is 1.
                    if i + 1 == prev_chem_symbol_start:
                        num_atoms = 1
                    else:
                        try:
                            num_atoms = int(formula[i + 1:prev_chem_symbol_start])
                        except ValueError:
                            raise ChemicalCompositionError(
                                'Badly-formed number of atoms: %s' % formula)

                    # Read isotope number if specified, else it is undefined (=0).
                    if formula[i] == ']':
                        brace_pos = formula.rfind('[', 0, i)
                        if brace_pos == -1:
                            raise ChemicalCompositionError(
                                'Badly-formed isotope number: %s' % formula)
                        try:
                            isotope_num = int(formula[brace_pos + 1:i])
                        except ValueError:
                            raise ChemicalCompositionError(
                                'Badly-formed isotope number: %s' % formula)
                        i = brace_pos - 1
                    else:
                        isotope_num = 0

                    # Match the element name to the mass_data.
                    element_found = False
                    # Sort the keys from longest to shortest to workaround
                    # the overlapping keys issue
                    for element_name in sorted(mass_data, key=len, reverse=True):
                        if formula.endswith(element_name, 0, i + 1):
                            isotope_string = _make_isotope_string(
                                element_name, isotope_num)
                            self[isotope_string] += num_atoms
                            i -= len(element_name)
                            prev_chem_symbol_start = i + 1
                            element_found = True
                            break

                    if not element_found:
                        raise ChemicalCompositionError(
                            'Unknown chemical element in the formula: %s' % formula)
            else:
                ch = formula[i]
                parse_stack += ch
                i -= 1
                if(ch == "("):
                    seek_mode -= 1
                    if seek_mode == 0:

                        resolve_stack.append(Composition(
                                             # Omit the last character, then reverse the parse
                                             # stack string.
                                             formula=parse_stack[:-1][::-1],
                                             mass_data=mass_data)
                                             * group_coef)
                        prev_chem_symbol_start = i + 1
                        seek_mode = False
                        parse_stack = ""
                elif(formula[i] == ")"):
                    seek_mode += 1
                else:
                    # continue to accumulate tokens
                    pass

        # Unspool the resolve stack, adding together the chunks
        # at this level. __add__ operates immutably, so must manually
        # loop through each chunk.
        for chunk in resolve_stack:
            for elem, cnt in chunk.items():
                self[elem] += cnt

    def _from_dict(self, comp):
        # for isotope_string, num_atoms in comp.items():
        #     element_name, isotope_num = _parse_isotope_string(
        #         isotope_string)

        #     # Remove explicitly undefined isotopes (e.g. X[0]).
        #     self[_make_isotope_string(element_name, isotope_num)] = (
        #         num_atoms)
        self.update(comp)

    def calc_mass(self, average=False, charge=None, mass_data=None):
        mdid = id(mass_data)
        if self._mass_args is not None and average is self._mass_args[0]\
                and charge == self._mass_args[1] and mdid == self._mass_args[2]:
            return self._mass
        else:
            self._mass_args = (average, charge, mdid)
            self._mass = pcalculate_mass(composition=self, average=average, charge=charge, mass_data=mass_data)
            return self._mass

    @property
    def mass(self):
        return self.calc_mass()

    def __init__(self, *args, **kwargs):
        """
        A Composition object stores a chemical composition of a
        substance. Basically it is a dict object, in which keys are the names
        of chemical elements and values contain integer numbers of
        corresponding atoms in a substance.

        The main improvement over dict is that Composition objects allow
        addition and subtraction.

        If ``formula`` is not specified, the constructor will look at the first
        positional argument and try to build the object from it. Without
        positional arguments, a Composition will be constructed directly from
        keyword arguments.

        Parameters
        ----------
        formula : str, optional
            A string with a chemical formula. All elements must be present in
            `mass_data`.
        mass_data : dict, optional
            A dict with the masses of chemical elements (the default
            value is :py:data:`nist_mass`). It is used for formulae parsing only.
        """
        defaultdict.__init__(self, int)
        mass_data = kwargs.get('mass_data', nist_mass)

        kw_sources = set(
            ('formula',))
        kw_given = kw_sources.intersection(kwargs)
        if len(kw_given) > 1:
            raise ChemicalCompositionError('Only one of {} can be specified!\n\
                Given: {}'.format(', '.join(kw_sources),
                                  ', '.join(kw_given)))

        elif kw_given:
            kwa = kw_given.pop()
            getattr(self, '_from_' + kwa)(kwargs[kwa], mass_data)
        # can't build from kwargs
        elif args:
            if isinstance(args[0], dict):
                self._from_dict(args[0])
            elif isinstance(args[0], str):
                try:
                    self._from_formula(args[0], mass_data)
                except ChemicalCompositionError:
                    raise ChemicalCompositionError(
                        'Could not create a Composition object from '
                        'string: "{}": not a valid sequence or '
                        'formula'.format(args[0]))
        else:
            self._from_dict(kwargs)

        self._mass = None
        self._mass_args = None


def pcalculate_mass(composition=None, formula=None, average=False, charge=None, mass_data=None):
    """Calculates the monoisotopic mass of a chemical formula or
    Composition object.

    Parameters
    ----------
    composition : Composition, optional
        A Composition object with the elemental composition of a substance.
    average : bool, optional
        If :py:const:`True` then the average mass is calculated. Note that mass
        is not averaged for elements with specified isotopes. Default is
        :py:const:`False`.
    charge : int, optional
        If not 0 then m/z is calculated: the mass is increased
        by the corresponding number of proton masses and divided
        by z.
    mass_data : dict, optional
        A dict with the masses of the chemical elements (the default
        value is :py:data:`nist_mass`).

    Returns
    -------
        mass : float
    """

    mass_data = mass_data or nist_mass
    # Make a deep copy of `composition` keyword argument.
    composition = composition if composition else PComposition(formula)

    # Get charge.
    old_charge = composition['H+']
    if charge is not None:
        if old_charge != 0:
            raise ChemicalCompositionError(
                'Charge is specified both by the number of protons and '
                '`charge` in kwargs')
        composition['H+'] = charge

    # Calculate mass.
    mass = 0.0
    for isotope_string in composition:
        element_name, isotope_num = _parse_isotope_string(isotope_string)
        # Calculate average mass if required and the isotope number is
        # not specified.
        if (not isotope_num) and average:
            for isotope in mass_data[element_name]:
                if isotope != 0:
                    mass += (composition[element_name]
                             * mass_data[element_name][isotope][0]
                             * mass_data[element_name][isotope][1])
        else:
            mass += (composition[isotope_string]
                     * mass_data[element_name][isotope_num][0])

    # Calculate m/z if required.
    if charge:
        mass /= abs(charge)
    if old_charge != 0:
        composition["H+"] = old_charge
    elif charge:
        del composition["H+"]
    return mass


# Checks to see if the Cython versions are available
if use_cython:  # pragma: no cover
    Composition = CComposition
    calculate_mass = ccalculate_mass
else:  # pragma: no cover
    Composition = PComposition
    calculate_mass = pcalculate_mass


def most_probable_isotopic_composition(*args, **kwargs):
    """Calculate the most probable isotopic composition of a
    chemical formula or |Composition| object.

    For each element, only two most abundant isotopes are considered.

    Parameters
    ----------
    formula : str, optional
        A string with a chemical formula.
    composition : |Composition|, optional
        A :py:class:`Composition` object with the elemental composition of a
        substance.
    elements_with_isotopes : list of str
        A list of elements to be considered in isotopic distribution
        (by default, every element has a isotopic distribution).
    mass_data : dict, optional
        A dict with the masses of chemical elements (.

    Returns
    -------
    out: tuple (|Composition|, float)
        A tuple with the most probable isotopic composition and its
        relative abundance.
    """

    composition = (dict(kwargs['composition']) if 'composition' in kwargs
                   else Composition(*args, **kwargs))

    # Removing isotopes from the composition.
    for isotope_string in composition:
        element_name, isotope_num = _parse_isotope_string(isotope_string)
        if isotope_num:
            composition[element_name] += composition.pop(isotope_string)

    mass_data = kwargs.get('mass_data', nist_mass)
    elements_with_isotopes = kwargs.get('elements_with_isotopes')
    isotopic_composition = Composition()

    for element_name in composition:
        if element_name == 'H+':
            continue
        if (not elements_with_isotopes or (element_name in elements_with_isotopes)):
            # Take the two most abundant isotopes.
            first_iso, second_iso = sorted(
                [(i[0], i[1][1]) for i in mass_data[element_name].items() if i[0]],
                key=lambda x: -x[1])[:2]

            # Write the number of isotopes of the most abundant type.
            first_iso_str = _make_isotope_string(element_name, first_iso[0])
            isotopic_composition[first_iso_str] = round(int(math.ceil(composition[element_name])) * first_iso[1])

            # Write the number of the second isotopes.
            second_iso_str = _make_isotope_string(element_name, second_iso[0])
            isotopic_composition[second_iso_str] = round(
                (composition[element_name] - isotopic_composition[first_iso_str]))
        else:
            isotopic_composition[element_name] = composition[element_name]

    return (isotopic_composition,
            isotopic_composition_abundance(
                composition=isotopic_composition,
                mass_data=mass_data))


def isotopic_composition_abundance(*args, **kwargs):
    """Calculate the relative abundance of a given isotopic composition
    of a molecule.

    Parameters
    ----------
    formula : str, optional
        A string with a chemical formula.
    composition : Composition, optional
        A Composition object with the isotopic composition of a substance.
    mass_data : dict, optional
        A dict with the masses of chemical elements (the default
        value is :py:data:`nist_mass`).

    Returns
    -------
    relative_abundance : float
        The relative abundance of a given isotopic composition.
    """

    composition = (Composition(kwargs['composition'])
                   if 'composition' in kwargs
                   else Composition(*args, **kwargs))

    isotopic_composition = defaultdict(dict)

    # Check if there are default and non-default isotopes of the same
    # element and rearrange the elements.
    for element in composition:
        if element == 'H+':
            continue
        element_name, isotope_num = _parse_isotope_string(element)

        # If there is already an entry for this element and either it
        # contains a default isotope or newly added isotope is default
        # then raise an exception.
        if ((element_name in isotopic_composition) and (isotope_num == 0 or 0 in isotopic_composition[element_name])):
            raise ChemicalCompositionError(
                'Please specify the isotopic states of all atoms of '
                '%s or do not specify them at all.' % element_name)
        else:
            isotopic_composition[element_name][isotope_num] = (
                composition[element])

    # Calculate relative abundance.
    mass_data = kwargs.get('mass_data', nist_mass)
    num1, num2, denom = 1., 1., 1.
    for element_name, isotope_dict in isotopic_composition.items():
        num1 *= math.factorial(sum(isotope_dict.values()))
        for isotope_num, isotope_content in isotope_dict.items():
            denom *= math.factorial(isotope_content)
            if isotope_num:
                num2 *= (mass_data[element_name][isotope_num][1] ** isotope_content)

    return num2 * (num1 / denom)


def formula(composition):
    keys = sorted(composition)
    return ''.join("%s%d" % (k, composition[k]) for k in keys)



'''
:class:`GlycanComposition`, :class:`MonosaccharideResidue`, and :class:`SubstituentResidue` are
useful for working with bag-of-residues where topology and connections are not relevant, but
the aggregate composition is known. These types work with a subset of the IUPAC three letter code
for specifying compositions.


>>> g = GlycanComposition(Hex=3, HexNAc=2)
>>> g["Hex"]
3
>>> r = MonosaccharideResidue.from_iupac_lite("Hex")
>>> r
MonosaccharideResidue(Hex)
>>> g[r]
3
>>> import glypy
>>> abs(g.mass() - glypy.motifs["N-Glycan core basic 1"].mass()) < 1e-5
True
>>> g2 = GlycanComposition(Hex=5)
>>> g["@n-acetyl"] = -2 # Remove two n-acetyl groups from the composition
>>> abs(g.mass() - g2.mass()) < 1e-5
True

'''

import warnings

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from glypy.utils import tree, uid
from glypy.utils.multimap import OrderedMultiMap

from glypy.composition import Composition
from glypy.structure.base import SaccharideCollection, MoleculeBase
from glypy.structure.glycan import Glycan
from glypy.structure.monosaccharide import Monosaccharide, ReducedEnd
from glypy.structure.substituent import Substituent
from glypy.structure.constants import (Anomer, Stem, Configuration, UnknownPosition)

from glypy.io import iupac
from glypy.io.iupac import (
    monosaccharide_reference as _monosaccharide_reference,
    resolve_special_base_type as _resolve_special_base_type,
    IUPACError)

from glypy.composition.base import formula
from glypy.composition.composition_transform import (
    derivatize, has_derivatization, strip_derivatization,
    _derivatize_reducing_end, _strip_derivatization_reducing_end,
    make_counter)

from six import string_types as basestring


monosaccharide_residue_reference = {}


class IUPACLiteMonosaccharideDeserializer(iupac.SimpleMonosaccharideDeserializer):

    def monosaccharide_from_iupac(self, monosaccharide_str, residue_class=None):
        """
        Parse a string in a limited subset of IUPAC three letter code into
        an instance of :class:`MonosaccharideResidue` or :class:`SubstituentResidue`.
        Parameters
        ----------
        monosaccharide_str: str
            The string to be parsed
        Returns
        -------
        MonosaccharideResidue
        """
        if residue_class is None:
            residue_class = MonosaccharideResidue
        try:
            match_dict = self.extract_pattern(monosaccharide_str)
            residue = self.build_residue(match_dict)
        except IUPACError:
            if monosaccharide_str.startswith(MolecularComposition.sigil):
                result = MolecularComposition.from_iupac_lite(monosaccharide_str)
                return result
            if monosaccharide_str == "Pent":
                warnings.warn("Pent is not proper IUPAC. Converting to Pen")
                return self.monosaccharide_from_iupac("Pen", residue_class)
            try:
                result = SubstituentResidue.from_iupac_lite(monosaccharide_str)
                return result
            except Exception:
                try:  # pragma: no cover
                    result = MolecularComposition.from_iupac_lite(monosaccharide_str)
                    return result
                except Exception:
                    raise IUPACError("Cannot find pattern in {}".format(monosaccharide_str))
        except TypeError:
            raise TypeError("Expected string, received {} ({})".format(monosaccharide_str, type(monosaccharide_str)))

        deriv = match_dict.get("derivatization", '')
        if deriv is not None and deriv != "":
            self.apply_derivatization(residue, deriv)
        return residue_class.from_monosaccharide(residue)

    def build_residue(self, match_dict):
        residue, linkage = super(IUPACLiteMonosaccharideDeserializer, self).build_residue(match_dict)
        return residue

    def __call__(self, string, residue_class=None):
        return self.monosaccharide_from_iupac(string, residue_class=residue_class)


class IUPACLiteMonosaccharideSerializer(iupac.SimpleMonosaccharideSerializer):
    def monosaccharide_to_iupac(self, residue):
        """
        Encode a subset of traits of a :class:`Monosaccharide`-like object
        using a limited subset of the IUPAC three letter code. The information
        present is sufficient to reconstruct a :class:`MonosaccharideResidue` instance
        reflecting the base type and its native substituents and modificats.
        .. note::
            This function is not suitable for use on whole |Glycan| objects. Instead,
            see :meth:`GlycanComposition.from_glycan` and :meth:`GlycanComposition.serialize`

        Parameters
        ----------
        residue: Monosaccharide
            The object to be encoded

        Returns
        -------
        str

        See Also
        --------
        :func:`from_iupac_lite`
        """
        try:
            string = super(IUPACLiteMonosaccharideSerializer, self).monosaccharide_to_iupac(residue)
        except (AttributeError, TypeError, ValueError):
            # if the residue passed was *really* a monosaccharide then this error is valid and
            # should propagate
            if isinstance(residue, Monosaccharide):
                raise
            else:
                string = str(residue)
        return string


from_iupac_lite = IUPACLiteMonosaccharideDeserializer()


to_iupac_lite = IUPACLiteMonosaccharideSerializer(
    iupac.monosaccharide_reference,
    iupac.SubstituentSerializer(monosaccharide_residue_reference))


Monosaccharide.register_serializer('iupac_lite', to_iupac_lite)


def drop_stem(residue, force=False):
    """Drops the stem, or the carbon ring stereochemical
    classification from this monosaccharide.

    Unless ``force`` is |True|, if :func:`~glypy.io.iupac.resolve_special_base_type`
    returns a truthy value, this function will do nothing.

    Parameters
    ----------
    residue : :class:`~.Monosaccharide`
        The monosaccharide to change
    force : bool, optional
        Whether or not to override known special case named monosaccharides

    Returns
    -------
    :class:`~.Monosaccharide`
        The mutated monosaccharide
    """
    if _resolve_special_base_type(residue) is None or force:
        residue.stem = (None,)
    return residue


def drop_positions(residue, force=False):
    """Drops the position classifiers from all links and modifications
    attached to this monosaccharide.

    Unless ``force`` is |True|, if :func:`~.iupac.resolve_special_base_type`
    returns a truthy value, this function will do nothing.

    Parameters
    ----------
    residue : :class:`~.Monosaccharide`
        The monosaccharide to change
    force : bool, optional
        Whether or not to override known special case named monosaccharides

    Returns
    -------
    :class:`~.Monosaccharide`
        The mutated monosaccharide
    """
    if _resolve_special_base_type(residue) is None or force:
        modifications = OrderedMultiMap()
        for _k, v in residue.modifications.items():
            modifications[UnknownPosition] = v
        residue.modifications = modifications

        for _p, link in list(residue.substituent_links.items()):
            link.break_link(refund=True)
            link.parent_position = UnknownPosition
            link.apply()
    return residue


def drop_configuration(residue, force=False):
    """Drops the absolute stereochemical configuration of this
    monosaccharide.

    Unless ``force`` is |True|, if :func:`~.iupac.resolve_special_base_type`
    returns a truthy value, this function will do nothing.

    Parameters
    ----------
    residue : :class:`~.Monosaccharide`
        The monosaccharide to change
    force : bool, optional
        Whether or not to override known special case named monosaccharides

    Returns
    -------
    :class:`~.Monosaccharide`
        The mutated monosaccharide
    """
    if _resolve_special_base_type(residue) is None or force:
        residue.configuration = (None,)
    return residue


water_composition = Composition({"O": 1, "H": 2})


class ResidueBase(object):
    __slots__ = ()

    def drop_stem(self, force=False):
        """Drops the stem, or the carbon ring stereochemical
        classification from this monosaccharide.

        Unless ``force`` is |True|, if :func:`~.iupac.resolve_special_base_type`
        returns a truthy value, this function will do nothing.

        Parameters
        ----------
        residue : :class:`~.Monosaccharide`
            The monosaccharide to change
        force : bool, optional
            Whether or not to override known special case named monosaccharides

        Returns
        -------
        :class:`~.Monosaccharide`
            The mutated monosaccharide
        """
        return self

    def drop_positions(self, force=False):
        """Drops the position classifiers from all links and modifications
        attached to this monosaccharide.

        Unless ``force`` is |True|, if :func:`~.iupac.resolve_special_base_type`
        returns a truthy value, this function will do nothing.

        Parameters
        ----------
        residue : :class:`~.Monosaccharide`
            The monosaccharide to change
        force : bool, optional
            Whether or not to override known special case named monosaccharides

        Returns
        -------
        :class:`~.Monosaccharide`
            The mutated monosaccharide
        """
        return self

    def drop_configuration(self, force=False):
        """Drops the absolute stereochemical configuration of this
        monosaccharide.

        Unless ``force`` is |True|, if :func:`~.iupac.resolve_special_base_type`
        returns a truthy value, this function will do nothing.

        Parameters
        ----------
        residue : :class:`~.Monosaccharide`
            The monosaccharide to change
        force : bool, optional
            Whether or not to override known special case named monosaccharides

        Returns
        -------
        :class:`~.Monosaccharide`
            The mutated monosaccharide
        """
        return self

    def to_iupac_lite(self):
        """Encode this residue using `iupac_lite` notation.

        Returns
        -------
        str
        """
        return to_iupac_lite(self)

    @classmethod
    def from_iupac_lite(cls, string):
        """Parse a string of `iupac_lite` notation to produce a residue object

        Parameters
        ----------
        string : :class:`str`
            The string to parse

        Returns
        -------
        ResidueBase
        """
        return from_iupac_lite(string, residue_class=cls)


class MonosaccharideResidue(Monosaccharide, ResidueBase):
    """Represents a :class:`Monosaccharide`-like object, save that it does
    not connect to other :class:`~.Monosaccharide` objects and does not have
    properties related to topology, specifically, :attr:`anomer`.

    A single :class:`MonosaccharideResidue` has lost a water molecule from its
    composition, reflecting its residual nature. This is accounted for when dealing
    with aggreates of residues. They also have altered carbon backbone occupancies.

    :class:`MonosaccharideResidue` objects are hashable and comparable on their
    `iupac_lite` representation, which is given by :meth:`__str__` or :meth:`name`.
    """
    __slots__ = ()

    @classmethod
    def from_monosaccharide(cls, monosaccharide, configuration=False, stem=True, ring=False):
        """Construct an instance of :class:`MonosaccharideResidue` from an instance
        of |Monosaccharide|. This function attempts to preserve derivatization if possible.

        This function will create a *deep copy* of `monosaccharide`.

        Parameters
        ----------
        monosaccharide : Monosaccharide
            The monosaccharide to be converted
        configuration : bool, optional
            Whether or not to preserve |Configuration|. Defaults to |False|
        stem : bool, optional
            Whether or not to preserve |Stem|. Defaults to |True|
        ring : bool, optional
            Whether or not to preserve |RingType|. Defaults to |False|

        Returns
        -------
        MonosaccharideResidue
        """
        residue = monosaccharide.clone(monosaccharide_type=cls)
        premass = residue.mass()

        deriv = has_derivatization(monosaccharide)
        strip_derivatization(residue)
        if _resolve_special_base_type(monosaccharide) is None:
            if not configuration:
                residue.configuration = (Configuration.x,)
            if not stem:
                residue.stem = (Stem.x,)
        if not ring:
            residue.ring_start = residue.ring_end = UnknownPosition
        if deriv:
            derivatize(residue, deriv)
        if residue.mass() != premass and not deriv:
            residue.composition += water_composition
        return residue

    def __init__(self, *args, **kwargs):
        super(MonosaccharideResidue, self).__init__(*args, **kwargs)
        self.composition -= water_composition
        self.anomer = Anomer.x

    def clone(self, *args, **kwargs):
        '''
        Copies just this |Monosaccharide| and its |Substituent| objects, creating a separate instance
        with the same data. All mutable data structures are duplicated and distinct from the original.

        Does not copy any :attr:`links` as this would cause recursive duplication of the entire |Glycan|
        graph.

        Parameters
        ----------
        prop_id: :class:`bool`
            Whether to copy :attr:`id` from ``self`` to the new instance
        fast: :class:`bool`
            Whether to use the fast-path initialization process in :meth:`MonosaccharideResidue.__init__`
        monosaccharide_type: :class:`type`
            A subclass of :class:`MonosaccharideResidue` to use

        Returns
        -------
        :class:`MonosaccharideResidue`

        '''
        kwargs.setdefault("monosaccharide_type", MonosaccharideResidue)
        residue = super(MonosaccharideResidue, self).clone(*args, **kwargs)
        return residue

    def __repr__(self):  # pragma: no cover
        return "MonosaccharideResidue(%s)" % self.name()

    def __str__(self):  # pragma: no cover
        return to_iupac_lite(self)

    def __hash__(self):  # pragma: no cover
        """Obtain a hash value from `self` based on :meth:`MonosaccharideResidue.name`.

        Returns
        -------
        int
        """
        return hash(self.name())

    def open_attachment_sites(self, max_occupancy=0):
        '''
        When attaching :class:`~.Monosaccharide` instances to other objects,
        bonds are formed between the carbohydrate backbone and the other object.
        If a site is already bound, the occupying object fills that space on the
        backbone and prevents other objects from binding there.

        Currently only cares about the availability of the hydroxyl group. As there
        is not a hydroxyl attached to the ring-ending carbon, that should not be
        considered an open site.

        If any existing attached units have unknown positions, we can't provide any
        known positions, in which case the list of open positions will be a :class:`list`
        of ``-1`` s of the length of open sites.

        A :class:`MonosaccharideResidue` has two fewer open attachment sites than
        the equivalent :class:`~.Monosaccharide`

        Parameters
        ----------
        max_occupancy: int
            The number of objects that may already be bound at a site before it
            is considered unavailable for attachment.

        Returns
        -------
        :class:`list`:
            The positions open for binding
        :class:`int`:
            The number of bound but unknown locations on the backbone.
        '''
        sites, unknowns = super(
            MonosaccharideResidue, self).open_attachment_sites(max_occupancy)
        return sites[:-2], unknowns

    def __eq__(self, other):
        '''
        Test for equality between :class:`MonosaccharideResidue` instances by comparing
        the result of :meth:`MonosaccharideResidue.name` calls between `self` and `other`.

        :meth:`MonosaccharideResidue.name` is an alias of :func:`to_iupac_lite` called on `self`
        '''
        if (other is None):
            return False
        if not isinstance(other, (MonosaccharideResidue, str)):
            return False
        return str(self) == str(other)

    def name(self):
        """Name this object according to `iupac_lite`.

        Returns
        -------
        str

        See Also
        --------
        :meth:`to_iupac_lite`
        """
        return to_iupac_lite(self)

    def residue_name(self):
        """Name this object according to `iupac_lite`, omitting any derivatization

        Returns
        -------
        str

        See Also
        --------
        :meth:`to_iupac_lite`
        :meth:`name`
        """
        name = self.name()
        return name.split("^")[0]

    drop_stem = drop_stem
    drop_positions = drop_positions
    drop_configuration = drop_configuration

    def copy_underivatized(self):
        """Create a copy of this residue without derivatization.

        Returns
        -------
        :class:`MonosaccharideResidue`
        """
        return from_iupac_lite.strip_derivatization(str(self), residue_class=self.__class__)

monosaccharide_residue_reference.update({
    k: MonosaccharideResidue.from_monosaccharide(v) for k, v in _monosaccharide_reference.items()
})


class FrozenMonosaccharideResidue(MonosaccharideResidue):
    '''
    A subclass of |MonosaccharideResidue| which caches the result of :func:`to_iupac_lite` and instances returned
    by :meth:`FrozenMonosaccharideResidue.clone` and :meth:`FrozenMonosaccharideResidue.from_iupac_lite`.
    Also treated as immutable after initialization through :meth:`FrozenMonosaccharideResidue.from_monosaccharide`.

    Note that directly calling :meth:`FrozenMonosaccharideResidue.from_monosaccharide` will not retrieve instances
    from the cache directly, and direct initialization using normal instance creation will neither touch the cache
    nor freeze the instance.

    This type is intended for use with :class:`FrozenGlycanComposition` to minimize the number of times
    :func:`from_iupac_lite` is called.
    '''
    __slots__ = ("_frozen", "_total_composition", "_hash", "_name", "_mass")

    _attribute_caching_slots = (
        '_total_composition', '_hash',
        '_mass'
    )

    # _frozen = False
    # _total_composition = None
    __cache = {}

    @classmethod
    def from_monosaccharide(cls, monosaccharide, *args, **kwargs):
        inst = super(FrozenMonosaccharideResidue, cls).from_monosaccharide(monosaccharide, *args, **kwargs)
        if str(inst) not in inst.get_cache():
            inst.get_cache()[str(inst)] = inst
            inst._frozen = True
        else:
            inst = inst.get_cache()[str(inst)]
        return inst

    def __init__(self, *args, **kwargs):
        self._total_composition = None
        self._mass = None
        # _name is left undefined to use a fast-path in __str__ by not testing for
        # presence first.
        # self._name = None
        # self._hash = None
        super(FrozenMonosaccharideResidue, self).__init__(*args, **kwargs)
        self._frozen = kwargs.get("_frozen", False)

    def __setattr__(self, key, value):
        try:
            is_frozen = self._frozen
        except AttributeError:
            is_frozen = False
        if is_frozen and key not in FrozenMonosaccharideResidue._attribute_caching_slots:
            self.get_cache().pop(str(self), None)
            raise FrozenError("Cannot change a frozen object")
        else:
            object.__setattr__(self, key, value)

    def __repr__(self):  # pragma: no cover
        return "FrozenMonosaccharideResidue(%s)" % self.name()

    def __hash__(self):  # pragma: no cover
        """Obtain a hash value from `self` based on :meth:`MonosaccharideResidue.name`.

        Returns
        -------
        int
        """
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(str(self))
            return self._hash

    def _update_hash(self):
        self._hash = hash(str(self))
        return self._hash

    def __eq__(self, other):
        '''
        Test for equality between :class:`MonosaccharideResidue` instances by comparing
        the result of :meth:`MonosaccharideResidue.name` calls between `self` and `other`.

        :meth:`MonosaccharideResidue.name` is an alias of :func:`to_iupac_lite` called on `self`
        '''
        if isinstance(other, MonosaccharideResidue):
            try:
                return self._name == other._name
            except AttributeError:
                return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        if (other is None):
            return False
        if not isinstance(other, (MonosaccharideResidue, str)):
            return False

    def _save_to_cache(self):
        self.get_cache()[str(self)] = self

    def __str__(self):
        try:
            return self._name
        except AttributeError:
            name = to_iupac_lite(self)
            self._name = name
            return name

    def clone(self, *args, **kwargs):
        '''
        Copies just this |Monosaccharide| and its |Substituent|s, creating a separate instance
        with the same data. All mutable data structures are duplicated and distinct from the original.

        Does not copy any :attr:`links` as this would cause recursive duplication of the entire |Glycan|
        graph.

        Parameters
        ----------
        prop_id: :class:`bool`
            Whether to copy :attr:`id` from ``self`` to the new instance
        fast: :class:`bool`
            Whether to use the fast-path initialization process in :meth:`Monosaccharide.__init__`
        monosaccharide_type: :class:`type`
            A subclass of :class:`Monosaccharide` to use

        Returns
        -------
        :class:`Monosaccharide`

        '''
        if self._frozen and kwargs.get(
                "monosaccharide_type",
                FrozenMonosaccharideResidue) is FrozenMonosaccharideResidue:
            return self
        else:
            return super(FrozenMonosaccharideResidue, self).clone(*args, **kwargs)

    def __getstate__(self):
        state = super(FrozenMonosaccharideResidue, self).__getstate__()
        state['_name'] = str(self)
        state['_total_composition'] = self.total_composition()
        return state

    def __setstate__(self, state):
        self._frozen = False
        self._total_composition = state.get('_total_composition')
        self._name = state.get('_name')
        self._hash = hash(str(self))
        super(FrozenMonosaccharideResidue, self).__setstate__(state)

    @classmethod
    def get_cache(cls):
        return cls.__cache

    @classmethod
    def from_iupac_lite(cls, string):
        cache = cls.get_cache()
        try:
            return cache[string]
        except KeyError:
            result = from_iupac_lite(string, residue_class=cls)
            if string not in cache:
                for k, v in cache.items():
                    if v == result:
                        cache[string] = v
                        break
                else:
                    cache[string] = result
            return result

    def total_composition(self):
        if self._frozen:
            if self._total_composition is None:
                self._total_composition = super(FrozenMonosaccharideResidue, self).total_composition()
                self._mass = None
            return self._total_composition
        else:
            return super(FrozenMonosaccharideResidue, self).total_composition()

    def mass(self, average=False, charge=0, mass_data=None, substituents=True):
        '''
        Calculates the total mass of ``self``.

        Parameters
        ----------
        average: bool, optional, defaults to False
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is ``charge``
        mass_data: dict, optional
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.
            Defaults to :const:`None`.
        substituents: bool, optional, defaults to True
            Whether or not to include substituents' masses.
        Returns
        -------
        :class:`float`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        if not average and charge == 0 and mass_data is None and self._frozen:
            if self._mass is None:
                self._mass = self.total_composition().calc_mass()
            return self._mass
        return self.total_composition().calc_mass(average=average, charge=charge, mass_data=mass_data)


class SubstituentResidue(Substituent, ResidueBase):
    r'''
    Represent substituent molecules unassociated with a specific
    monosaccharide residue.


    .. note::
        :class:`SubstituentResidue`'s composition value includes the losses for forming a bond between
        a monosaccharide residue and the substituent.

    Attributes
    ----------
    name: str
        As in |Substituent|, but with :attr:`SubstituentResidue.sigil` prepended.
    composition: |Composition|
    links: |OrderedMultiMap|
    _order: |int|
    '''
    #: All substituent string identifiers are prefixed with this character
    #: for the :func:`from_iupac_lite` parser
    sigil = "@"

    def __init__(self, name, composition=None, id=None, links=None,
                 can_nh_derivatize=None, is_nh_derivatizable=None, derivatize=False,
                 attachment_composition=None):
        if name.startswith(SubstituentResidue.sigil):
            name = name[1:]
        elif name.startswith(MolecularComposition.sigil):
            raise TypeError("Invalid Sigil. SubstituentResidue instances must be given names with either"
                            " no sigil prefix or with '@'")
        super(SubstituentResidue, self).__init__(
            name=name, composition=composition, links=links, id=id,
            can_nh_derivatize=can_nh_derivatize, is_nh_derivatizable=is_nh_derivatizable,
            derivatize=derivatize, attachment_composition=attachment_composition)

        self._residue_name = SubstituentResidue.sigil + self._name
        self.composition -= self.attachment_composition
        self.composition -= {"H": 1}
        self._hash = None

    def __hash__(self):  # pragma: no cover
        """Obtain a hash value from `self` based on :attr:`name`.

        Returns
        -------
        int
        """
        try:
            if self._hash is None:
                self._hash = hash(self._residue_name)
            return self._hash
        except AttributeError:
            return hash(self._residue_name)

    def __getstate__(self):
        state = super(SubstituentResidue, self).__getstate__()
        state['_residue_name'] = self._residue_name
        return state

    def __setstate__(self, state):
        super(SubstituentResidue, self).__setstate__(state)
        self._residue_name = state.get("_residue_name")

    def to_iupac_lite(self):
        return self._residue_name

    __str__ = to_iupac_lite

    def __repr__(self):  # pragma: no cover
        return "SubstituentResidue(%s)" % self._residue_name

    @classmethod
    def from_iupac_lite(cls, name):
        return cls(name)

    def __eq__(self, other):
        if (other is None):
            return False
        if isinstance(other, str):
            return other == self._residue_name
        if not isinstance(other, SubstituentResidue):
            return False
        return self.name == other.name

    def __ne__(self, other):  # pragma: no cover
        return not self == other

    def _backsolve_original_composition(self):
        comp = super(SubstituentResidue, self)._backsolve_original_composition()
        comp += {"H": 1}
        return comp

    def copy_underivatized(self):
        inst = self.clone()
        strip_derivatization(inst)
        return inst


class MolecularComposition(MoleculeBase, ResidueBase):  # pragma: no cover
    sigil = "#"

    __slots__ = ('name', 'composition', '_hash')

    def __init__(self, name, composition):
        self.name = name
        self.composition = composition
        self._hash = None

    def mass(self, average=False, charge=0, mass_data=None):
        return self.composition.calc_mass(average=average, charge=charge, mass_data=mass_data)

    def __repr__(self):
        return "%s%s%s%s" % (
            self.sigil, self.name, self.sigil,
            formula(self.composition))

    to_iupac_lite = __repr__

    def __reduce__(self):
        return self.__class__, (self.name, self.composition)

    def open_attachment_sites(self, *args, **kwargs):
        return 0

    def clone(self):
        return self.__class__(self.name, Composition(self.composition))

    def total_composition(self):
        return self.composition.clone()

    @classmethod
    def from_iupac_lite(cls, string):
        if not string.startswith(cls.sigil):
            raise TypeError("%s does not start with header %s" % (string, cls.sigil))
        _, header, composition = string.split("#")
        name = header
        return cls(name, Composition(composition))

    def __hash__(self):  # pragma: no cover
        """Obtain a hash value from `self` based on :attr:`name`.

        Returns
        -------
        int
        """
        try:
            if self._hash is None:
                self._hash = hash(self.name)
            return self._hash
        except AttributeError:
            return hash(self.name)

    def __eq__(self, other):
        try:
            return self.name == other or self.name == other.name
        except AttributeError:
            return self.name == str(other)

    def __ne__(self, other):
        return not (self == other)


class _CompositionBase(dict):

    def _setitem_fast(self, key, value):
        dict.__setitem__(self, key, value)

    def _getitem_fast(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            return 0

    def __reduce__(self):
        return self.__class__, (), self.__getstate__()

    def __getstate__(self):
        d = {
            'mapping': dict(self),
            'reducing_end': self._reducing_end,
            'composition_offset': self._composition_offset
        }
        return d

    def __setstate__(self, state):
        self.update(state['mapping'])
        self._reducing_end = state['reducing_end']
        self._composition_offset = state['composition_offset']

    @classmethod
    def _empty(cls):
        inst = cls.__new__(cls)
        inst._composition_offset = water_composition.clone()
        inst._reducing_end = None
        inst._mass = None
        return inst

    def _update_from_typed_map(self, template, copy_nodes=False):
        if copy_nodes:
            for name, count in template.items():
                self._setitem_fast(name.clone(), count)
        else:
            for name, count in template.items():
                self._setitem_fast(name, count)
        reduced = template.reducing_end
        if reduced is not None:
            self.reducing_end = reduced.clone()
        self._mass = None

    def serialize(self):
        """Convert a glycan composition into a curly brace-enclosed string specifying
        pairs of `iupac_lite` and a integer count.

        If the glycan is reduced, it will be appended to the closing brace following a
        `$` character.

        Returns
        -------
        str
        """
        form = "{%s}" % '; '.join("{}:{}".format(str(k), v) for k, v in sorted(
            self.items(), key=lambda x: (x[0].mass(), str(x[0]))) if v != 0)
        reduced = self._reducing_end
        if reduced is not None:
            form = "%s$%s" % (form, formula(reduced.total_composition()))
        return form

try:
    from glypy._c.structure.glycan_composition import _CompositionBase
except ImportError:
    pass


class GlycanComposition(_CompositionBase, SaccharideCollection):
    """
    Describe a glycan  as a collection of :class:`MonosaccharideResidue` counts without
    explicit linkage information relating how each monosaccharide is connected to its neighbors.

    This class subclasses |dict|, and assumes that keys will either be :class:`MonosaccharideResidue`
    instances, :class:`SubstituentResidue` instances, or strings in `iupac_lite` format which will be parsed
    into one of these types. While other types may be used, this is not recommended. All standard |dict| methods
    are supported.

    |GlycanComposition| objects may be derivatized just as |Glycan| objects are, with
    :func:`glypy.composition.composition_transform.derivatize` and
    :func:`glypy.composition.composition_transform.strip_derivatization`.

    GlycanComposition objects also support composition arithmetic, and can be added or subtracted from each other
    or multiplied by an integer.

    As GlycanComposition is not a complete structure, they cannot be translated into text formats as
    full |Glycan| objects are. They may instead be converted to and from a short-form text notation using
    :meth:`GlycanComposition.serialize` and reconstructed from this format using :meth:`GlycanComposition.parse`.

    Attributes
    ----------
    reducing_end : ReducedEnd
        Describe the reducing end of the aggregate without binding it to a specific monosaccharide.
        This will contribute to composition and mass calculations.
    _composition_offset: CComposition
        Account for the one water molecule's worth of composition left over from applying the "residue"
        transformation to each monosaccharide in the aggregate.
    """
    _monosaccharide_type = MonosaccharideResidue

    _key_parser = staticmethod(from_iupac_lite)

    @classmethod
    def from_glycan(cls, glycan):
        """
        Convert a |Glycan| into a |GlycanComposition|.

        Parameters
        ----------
        glycan : :class:`~.Glycan`
            The instance to be converted

        Returns
        -------
        GlycanComposition
        """
        inst = cls()
        glycan = tree(glycan)
        inst.extend(glycan)
        inst.reducing_end = glycan.reducing_end
        deriv = has_derivatization(glycan.root)
        if deriv:
            inst._composition_offset += (
                deriv.total_composition() - deriv.attachment_composition_loss()) * 2
        return inst

    def __init__(self, *args, **kwargs):
        """Initialize a :class:`GlycanComposition` using the provided objects or keyword
        arguments, imitating the :class:`dict` initialization signature.

        If a :class:`Mapping` is provided as a positional argument, it will be used as a
        template. If arbitrary keyword arguments are provided, they will be interpreted
        using :meth:`update`. As a special case, if another :class:`GlycanComposition` is
        provided, its :attr:`reducing_end` attribute will also be copied.

        Parameters
        ----------
        *args:
            Arbitrary positional arguments
        **kwargs:
            Arbitrary keyword arguments
        """
        # dict.__init__ just calls C update method. Expensive parameter parsing
        # _CompositionBase.__init__(self)
        self._reducing_end = None
        self._mass = None
        if args or kwargs:
            self.update(*args, **kwargs)
        if args:
            template = args[0]
            if isinstance(template, GlycanComposition):
                reduced = template.reducing_end
                if reduced is not None:
                    self.reducing_end = reduced.clone()
                self._composition_offset = template._composition_offset.clone()
            else:
                self._composition_offset = water_composition.clone()
        else:
            self._composition_offset = water_composition.clone()

    def __setitem__(self, key, value):
        """
        Set the quantity of `key` to `value`

        If `key` is a string, it will be passed through :func:`from_iupac_lite`

        If `key` has a reducing end value, that reducing end will be set on `self`

        Parameters
        ----------
        key : str, MonosaccharideResidue, SubstituentResidue, or MolecularComposition
            The entity to store
        value : int
            The value to store
        """
        if isinstance(key, basestring):
            key = self._key_parser(key)
        if key.node_type is Monosaccharide.node_type and key.reducing_end is not None:
            self.reducing_end = key.reducing_end
            key = key.clone()
            key.reducing_end = None
        _CompositionBase.__setitem__(self, key, int(value))
        self._mass = None

    def __getitem__(self, key):
        """
        Get the quantity of `key`

        If `key` is a string, it will be passed through :func:`from_iupac_lite`

        If `key` has a reducing end value, that reducing end will be set on `self`

        Parameters
        ----------
        key : str, MonosaccharideResidue, SubstituentResidue, or MolecularComposition
            The entity to store

        Returns
        -------
        int
        """
        if isinstance(key, basestring):
            key = self._key_parser(key)
        try:
            return _CompositionBase.__getitem__(self, key)
        except KeyError:
            return 0

    def __delitem__(self, key):
        if isinstance(key, basestring):
            key = self._key_parser(key)
        _CompositionBase.__delitem__(self, key)
        self._mass = None

    def mass(self, average=False, charge=0, mass_data=None):
        '''
        Calculates the total mass of ``self``.

        Parameters
        ----------
        average: bool, optional, defaults to False
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is ``charge``
        mass_data: dict, optional
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.
            Defaults to :const:`None`.

        Returns
        -------
        :class:`float`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        if self._mass is not None and charge == 0:
            return self._mass
        if charge == 0:
            mass = self._composition_offset.mass
            for residue_type, count in self.items():
                mass += residue_type.mass(average=average, charge=0, mass_data=mass_data) * count
            if self._reducing_end is not None:
                mass += self._reducing_end.mass(average=average, charge=0, mass_data=mass_data)
            self._mass = mass
        else:
            mass = self.total_composition().calc_mass(average=average, charge=charge, mass_data=mass_data)
        return mass

    def update(self, *args, **kwargs):
        if len(args) == 1:
            if isinstance(args[0], Mapping):
                for name, count in args[0].items():
                    if count != 0:
                        self[name] = count
            else:
                for name, count in args:
                    if count != 0:
                        self[name] = count
        for name, count in kwargs.items():
            if count != 0:
                self[name] = count
        self._mass = None

    def extend(self, *args):
        if not isinstance(args[0], MonosaccharideResidue):
            if isinstance(args[0], (Monosaccharide)):
                args = map(MonosaccharideResidue.from_monosaccharide, args)
            elif isinstance(args[0], Glycan):
                args = map(
                    MonosaccharideResidue.from_monosaccharide,
                    [node for node in args[0] if node.node_type is MonosaccharideResidue.node_type])
            else:
                raise TypeError(
                    "Can't convert {} to MonosaccharideResidue".format(
                        type(args[0])))
        for residue in args:
            self[residue] += 1

    def __iadd__(self, other):
        for elem, cnt in (other.items()):
            self[elem] += cnt
        return self

    def __add__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] += cnt
        return result

    def __radd__(self, other):
        return self + other

    def __isub__(self, other):
        for elem, cnt in other.items():
            self[elem] -= cnt
        return self

    def __sub__(self, other):
        result = self.clone()
        for elem, cnt in other.items():
            result[elem] -= cnt
        return result

    def __rsub__(self, other):
        return (self - other) * (-1)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError(
                'Cannot multiply Composition by non-integer',
                other)
        prod = {}
        for k, v in self.items():
            prod[k] = v * other

        return self.__class__(prod)

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        if isinstance(other, basestring):
            return str(self) == other
        if not isinstance(other, Mapping):
            return False
        self_items = set([i for i in self.items() if i[1]])
        other_items = set([i for i in other.items() if i[1]])
        return self_items == other_items

    def __ne__(self, other):
        return not (self == other)

    def __neg__(self):
        return -1 * self

    def __missing__(self, key):
        return 0

    def __contains__(self, key):
        if isinstance(key, basestring):
            key = self._key_parser(key)
        return _CompositionBase.__contains__(self, key)

    def drop_stems(self):
        for t in self:
            drop_stem(t)
        self.collapse()
        return self

    def drop_positions(self):
        for t in self:
            drop_positions(t)
        self.collapse()
        return self

    def drop_configurations(self):
        for t in self:
            drop_configuration(t)
        self.collapse()
        return self

    def total_composition(self):
        '''
        Computes the sum of the composition of all |Monosaccharide| objects in ``self``

        Returns
        -------
        :class:`~glypy.composition.Composition`
        '''

        comp = self._composition_offset.clone()
        for residue, count in self.items():
            comp += residue.total_composition() * count
        if self._reducing_end is not None:
            comp += self._reducing_end.total_composition()
        return comp

    def collapse(self):
        '''
        Merge redundant keys.

        After performing a structure-detail removing operation like
        :meth:`drop_positions`, :meth:`drop_configurations`, or :meth:`drop_stems`,
        monosaccharide keys may be redundant.

        `collapse` will merge keys which refer to the same type of molecule.
        '''
        items = list(self.items())
        self.clear()
        for k, v in items:
            self[k] += v
        return self

    def query(self, query, exact=True, **kwargs):
        """Return the total count of all residues in `self` which
        match `query` using :func:`glypy.io.nomenclature.identity.is_a`

        Parameters
        ----------
        query : :class:`~.MonosaccharideResidue` or :class:`str`
            A monosaccharide residue or a string which will be converted into one by
            :func:`from_iupac_lite` to test for an `is-a` relationship with.
        exact : bool, optional
            Passed to :func:`~.is_a`. Explicitly |True| by default
        **kwargs
            Passed to :func:`~.is_a`

        Returns
        -------
        int
            The total count of all residues which satisfy the `is-a` relationship

        See Also
        --------
        :func:`glypy.io.nomenclature.identity.is_a`

        """
        from glypy.io.nomenclature.identity import is_a
        if isinstance(query, basestring):
            query = self._key_parser(query)
        count = 0
        for key, value in self.items():
            if is_a(key, query, exact=exact, **kwargs):
                count += value
        return count

    def reinterpret(self, references, exact=True, **kwargs):
        """Aggregate the counts of all residues in `self` for each
        monosaccharide in `references` satisfying an `is-a` relationship,
        collapsing multiple residues to a single key. Any residue not
        aggregated will be preserved as-is.

        .. note::
            The order of ``references`` matters as any residue matched by
            a reference will not be considered for later references.

        Parameters
        ----------
        references : :class:`Iterable` of :class:`~.MonosaccharideResidue`
            The monosaccharides with which to test for an `is-a` relationship
        exact : bool, optional
            Passed to :func:`~.is_a`. Explicitly |True| by default
        **kwargs
            Passed to :func:`~.is_a`

        Returns
        -------
        :class:`~.GlycanComposition`
            self after key collection and collapse
        """
        from glypy.io.nomenclature.identity import is_a
        new_counts = []
        pairs = list(self.items())
        remaining_pairs = []
        for ref in references:
            count = 0
            for key, value in pairs:
                if is_a(key, ref, exact=exact, **kwargs):
                    count += value
                else:
                    remaining_pairs.append((key, value))
            if count > 0:
                new_counts.append((ref, count))
            pairs = remaining_pairs
            remaining_pairs = []
        self.clear()
        for key, value in new_counts:
            self[key] = value
        for key, value in pairs:
            self[key] = value
        return self

    @property
    def reducing_end(self):
        return self._reducing_end

    @reducing_end.setter
    def reducing_end(self, value):
        self._invalidate()
        self._reducing_end = value

    def set_reducing_end(self, value):
        self._invalidate()
        self._reducing_end = value

    def _invalidate(self):
        self._mass = None

    @property
    def composition_offset(self):
        return self._composition_offset

    @composition_offset.setter
    def composition_offset(self, value):
        self._invalidate()
        self._composition_offset = value


    def clone(self, propogate_composition_offset=True, copy_nodes=True):
        dup = self._empty()
        dup._update_from_typed_map(self, copy_nodes=copy_nodes)
        if not propogate_composition_offset:
            dup._composition_offset = Composition('H2O')
        else:
            dup._composition_offset = self._composition_offset.clone()
        return dup

    # inheriting from dict overwrites MoleculeBase.copy
    def copy(self, *args, **kwargs):
        return self.clone(*args, **kwargs)

    def __str__(self):
        return self.serialize()

    @classmethod
    def _get_parse_tokens(cls, string):
        string = str(string)
        parts = string.split('$')
        if len(parts) == 1:
            tokens = parts[0]
            reduced = None
        elif len(parts) == 2:
            tokens, reduced = parts
        else:
            raise ValueError("Could not interpret %r" % string)
        tokens = tokens[1:-1].split('; ')
        return tokens, reduced

    def _handle_reduction_and_derivatization(self, reduced, deriv):
        if reduced:
            reduced = ReducedEnd(Composition(reduced))
            self.reducing_end = reduced
        if deriv:
            self._derivatized(deriv.clone(), make_counter(uid()), include_reducing_end=False)

    @classmethod
    def parse(cls, string):
        """Parse a :class:`str` into a :class:`GlycanComposition`.

        This will parse the format produced by :meth:`serialize`

        Parameters
        ----------
        string : :class:`str`
            The string to parse

        Returns
        -------
        :class:`GlycanComposition`
        """
        tokens, reduced = cls._get_parse_tokens(string)
        inst = cls._empty()
        deriv = None
        for token in tokens:
            try:
                residue, count = _parse_name_count(token)
            except ValueError:
                if string == "{}":
                    return inst
                else:
                    raise ValueError("Malformed Token, %s" % (token,))
            key = cls._key_parser(residue)
            if "^" in residue:
                _deriv = has_derivatization(key)
                if _deriv:
                    deriv = _deriv
            inst._setitem_fast(key, count)
        inst._handle_reduction_and_derivatization(reduced, deriv)
        return inst

    def _derivatized(self, substituent, id_base, include_reducing_end=True):
        n = 2
        items = list(self.items())
        self.clear()
        for k, v in items:
            self._setitem_fast(k, v)
            if k.node_type is Substituent.node_type:
                n -= v
        self._composition_offset += (
            substituent.total_composition() -
            substituent.attachment_composition_loss() * 2) * n
        if self._reducing_end is not None and include_reducing_end:
            _derivatize_reducing_end(self._reducing_end, substituent, id_base)
        self.collapse()
        self._invalidate()

    def _strip_derivatization(self):
        self._composition_offset = Composition("H2O")
        if self._reducing_end is not None:
            _strip_derivatization_reducing_end(self._reducing_end)
        self.collapse()
        self._invalidate()

    def _invalidate(self):
        self._mass = None


from_glycan = GlycanComposition.from_glycan
parse = GlycanComposition.parse


class FrozenGlycanComposition(GlycanComposition):
    '''
    A subclass of |GlycanComposition| which uses :class:`FrozenMonosaccharideResidue` instead
    of |MonosaccharideResidue| which reduces the number of times :func:`from_iupac_lite` is called.

    Only use this type if residue names are pre-validated, residue types will not be transformed,
    and when creating many, many instances. :func:`from_iupac_lite` invokes expensive introspection
    algorithms which can be costly when repeatedly manipulating the same residue types.
    '''

    _str = None

    _monosaccharide_type = FrozenMonosaccharideResidue

    _key_parser = staticmethod(FrozenMonosaccharideResidue.from_iupac_lite)

    def __setitem__(self, key, value):
        key = self._key_parser(str(key))
        _CompositionBase.__setitem__(self, key, value)
        self._invalidate()

    def __getitem__(self, key):
        if not isinstance(key, FrozenMonosaccharideResidue):
            key = self._key_parser(str(key))
        return _CompositionBase.__getitem__(self, key)

    def __delitem__(self, key):
        key = self._key_parser(str(key))
        _CompositionBase.__delitem__(self, key)
        self._invalidate()

    @classmethod
    def parse(cls, string):
        tokens, reduced = cls._get_parse_tokens(string)
        inst = cls._empty()
        deriv = None
        key_parser = cls._key_parser
        for token in tokens:
            try:
                residue, count = _parse_name_count(token)
            except ValueError:
                if string == "{}":
                    return inst
                else:
                    raise ValueError("Malformed Token, %s" % (token,))
            key = key_parser(residue)
            if "^" in residue:
                _deriv = has_derivatization(key)
                if _deriv:
                    deriv = _deriv
            inst._setitem_fast(key, count)
        inst._handle_reduction_and_derivatization(reduced, deriv)
        return inst

    def serialize(self):
        if self._str is None:
            self._str = super(FrozenGlycanComposition, self).serialize()
        return self._str

    __str__ = serialize

    def __contains__(self, key):
        if isinstance(key, basestring):
            key = self._key_parser(key)
        return _CompositionBase.__contains__(self, key)

    def thaw(self):
        """Convert this :class:`FrozenGlycanComposition` into a :class:`GlycanComposition`
        that is not frozen.

        Returns
        -------
        :class:`GlycanComposition`
        """
        return GlycanComposition.parse(self)

    def extend(self, *args):
        if not isinstance(args[0], FrozenMonosaccharideResidue):
            if isinstance(args[0], (Monosaccharide)):
                args = map(FrozenMonosaccharideResidue.from_monosaccharide, args)
            elif isinstance(args[0], Glycan):
                args = map(
                    FrozenMonosaccharideResidue.from_monosaccharide,
                    [node for node in args[0]
                     if node.node_type is FrozenMonosaccharideResidue.node_type])
            else:
                raise TypeError(
                    "Can't convert {} to FrozenMonosaccharideResidue".format(
                        type(args[0])))
        for residue in args:
            self[residue] += 1

    def _validate(self):
        '''Populate the caching fields used for common behaviors, e.g.
        mass and string representation.
        '''
        if self._mass is None:
            self.mass()
        if self._str is None:
            self.serialize()

    def _invalidate(self):
        '''Clear the caching fields, forcing them to all be recalculated
        when next requested.
        '''
        self._mass = None
        self._str = None
        self._total_composition = None

    def clone(self, propogate_composition_offset=True, copy_nodes=False):
        dup = self._empty()
        dup._update_from_typed_map(self, copy_nodes=copy_nodes)
        if not propogate_composition_offset:
            dup._composition_offset = Composition('H2O')
        else:
            dup._composition_offset = self._composition_offset.clone()
        return dup


class FrozenError(ValueError):
    pass


class HashableGlycanComposition(FrozenGlycanComposition):
    def __str__(self):
        self._validate()
        # Directly use internal cache variable to save time calling
        # the super method chain
        return self._str

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, HashableGlycanComposition):
            if self._str is not None:
                if other._str is not None:
                    return self._str == other._str
                return self._str == str(other)
            return str(other) == str(self)
        else:
            return super(HashableGlycanComposition, self).__eq__(other)


def _parse_name_count(string):
    name, count = string.split(":")
    count = int(count)
    return name, count


try:
    _has_c = True
    from glypy._c.utils import get_parse_tokens, parse_name_count as _parse_name_count
    GlycanComposition._get_parse_tokens = get_parse_tokens
except ImportError:
    _has_c = False



import unittest

from glypy.utils import lazy
from glypy.structure.named_structures import MonosaccharideIndex


class ProxyTest(unittest.TestCase):
    def test_init(self):
        index = lazy.ProxyObject(MonosaccharideIndex)
        self.assertEqual(index._initializer, MonosaccharideIndex)
        self.assertEqual(index._source, None)
        fucose = index.Fucose
        self.assertTrue(isinstance(index._source, MonosaccharideIndex))
        index.dHex = fucose
        self.assertEqual(index.dHex, index.Fucose)
        index['ldeoxyGal'] = fucose
        self.assertEqual(index['ldeoxyGal'], index.Fucose)



from glypy.structure import Stem, Modification
from glypy.utils.enum import Enum
from .cfg_symbols import CFGNomenclature


class SNFGNomenclature(CFGNomenclature):
    class ResidueColor(Enum):
        white = "#ffffff"
        blue = "#0180ff"
        green = "#01A650"
        yellow = "#ffd706"
        light_blue = "#91cce8"
        pink = "#fe87c5"
        purple = "#a21fff"
        brown = "#8e6319"
        orange = "#f37922"
        red = "#fc0200"
        generic = 'white'

    stem_to_color = {
        Stem.glc: ResidueColor.blue,
        Stem.man: ResidueColor.green,
        Stem.gal: ResidueColor.yellow,
        Stem.gul: ResidueColor.orange,
        Stem.alt: ResidueColor.pink,
        Stem.all: ResidueColor.purple,
        Stem.tal: ResidueColor.light_blue,
        Stem.ido: ResidueColor.brown,
        Stem.xyl: ResidueColor.orange,
        Stem.lyx: ResidueColor.yellow,
        Stem.x: ResidueColor.generic,
    }

    def residue_color(self, monosaccharide):
        '''
        Determine which color to use to represent `monosaccharide` under the CFG
        symbol nomenclature.

        Parameters
        ----------
        monosaccharide: |Monosaccharide|
            The residue to be rendered

        Returns
        -------
        ResidueColor.EnumValue

        '''
        if any(mod == Modification.a for p, mod in monosaccharide.modifications.items()):
            return self.resolve_acid_color(monosaccharide)
        if "hex" in [monosaccharide.superclass]:
            if any(mod == Modification.d for p, mod in monosaccharide.modifications.items()) and\
                    monosaccharide.stem == (Stem.gal,):
                return self.ResidueColor.red
        try:
            return self.stem_to_color[monosaccharide.stem[0]]
        except KeyError:
            return self.ResidueColor.generic

    def resolve_acid_color(self, monosaccharide):
        '''
        Resolve the special case in :func:`residue_color` for acidic residues
        '''
        if ('gro' in monosaccharide.stem) and ('gal' in monosaccharide.stem):
            if any(sub.name == 'n_acetyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueColor.purple
            elif any(sub.name == 'n_glycolyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueColor.light_blue
            elif any(sub.name == 'amino' for p, sub in monosaccharide.substituents()):
                return self.ResidueColor.brown
            else:
                return self.ResidueColor.green
        else:
            return self.stem_to_color[monosaccharide.stem[0]]


def test_color_pallette(ax):
    d = SNFGNomenclature()
    for i, kv in enumerate(SNFGNomenclature.ResidueColor):
        name, color = kv[0], kv[1]
        if name == "?":
            print(kv)
        d.draw_square(ax, i, 0, color)
        ax.text(i, -0.5, name, ha='center')
    ax.set_xlim(-1, i + 1)
    ax.set_ylim(-1, 1)



import os
import sqlite3
import logging
import functools
try:
    from collections import Counter
    from collections.abc import Iterable, Callable
except ImportError:
    from collections import Counter, Iterable, Callable


import glypy

from glypy.utils import pickle, classproperty, make_struct
from glypy.io.nomenclature.identity import naive_name_monosaccharide
from glypy.algorithms import subtree_search

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

Taxon = make_struct("Taxon", ("tax_id", "name", 'entries'))
Aglyca = make_struct("Aglyca", ("name", "reducing", 'entries'))
DatabaseEntry = make_struct("DatabaseEntry", ("database", "id"))
Motif = make_struct("Motif", ("name", "id", "motif_class"))


def _resolve_column_data_mro(cls):
    '''
    Given a class with :attr:`__column_data_map` mangled attributes
    from its class hierarchy, extract in descending order each
    `dict`, overwriting old settings as it approaches the most
    recently descended class.

    Parameters
    ----------
    cls: type
        The type to attempt to extract column_data mappings for
        along the MRO

    Returns
    -------
    dict:
        The column_data mapping describing the entire class hierarchy along
        `cls`'s MRO.
    '''
    attrs = []
    for attr in dir(cls):
        if "column_data_map" in attr:
            attrs.append(attr)
    mapping = {attr[1:].split("__column_data_map")[0]: attr for attr in attrs}
    meta_map = {}
    for typ in cls.mro()[::-1][1:]:
        column_data = mapping.get(typ.__name__, "__column_data_map")
        meta_map.update(getattr(cls, column_data, {}))
    return meta_map


def _extract_querymethods(cls):
    methods = {}
    for name, value in cls.__dict__.items():
        if isinstance(value, QueryMethod):
            methods[name] = value
    return methods


def _resolve_querymethods_mro(cls):
    methods = {}
    for typ in cls.__mro__[::-1]:
        methods.update(_extract_querymethods(typ))
    return methods


def column_data(name, dtype, transform):
    '''
    Decorator for adding a new column to the SQL table mapped to a record class

    Parameters
    ----------
    name: str
        Name of the new column
    dtype: str
        The SQL data type to encode the column as
    transform: function
        The function to extract the value of the column from a record

    Returns
    -------
    function:
        Decorator that will call :meth:`.add_column_data` with `name`, `dtype` and
        `transform` on the decorated class after instantiation
    '''
    def func(cls):
        cls.add_column_data(name, dtype, transform)
        return cls
    return func


class QueryMethod(object):
    def __init__(self, func=None):
        self.func = func

    def bind(self, obj):
        """Binds database parameters to the wrapped function

        Parameters
        ----------
        obj: RecordDatabase

        Returns
        -------
        function: The partially parameterized function bound to `obj`
        """
        @functools.wraps(self.func)
        def forward_binding(*args, **kwargs):
            return self.func(obj.record_type, obj, *args, **kwargs)
        return forward_binding

    def __call__(self, record_type, conn, *args, **kwargs):
            return self.func(record_type, conn, *args, **kwargs)


def querymethod(func):
    """Decorator for creating patching methods onto databases

    Parameters
    ----------
    func : method
        Method to be bound

    Returns
    -------
    QueryMethod
    """
    return QueryMethod(func)


class GlycanRecordBase(object):
    '''
    Defines the base class for SQL serialize-able records carrying
    glycan structure data and column_data. Includes tools for extending the
    SQL schema describing the structure to make new information query-able.

    .. warning::
        This is a base class. All user-defined record classes should be based on :class:`.GlycanRecord`

    Intended for storage in a :class:`RecordDatabase` instance, a wrapper around
    an Sqlite3 database file

    Attributes
    ----------
    structure: |Glycan|
        The |Glycan| structure being described
    motifs: |dict|
        A collection of motif names to motif locations
    dbxref: |dict|
        A collection of database names to database IDs for this structure
    aglycones: |dict|
        A collection of aglycone names to structures
    taxa: |dict|
        A collection of taxon ids to taxon objects
    id: |int|
        A unique numeric identifier used as a primary key

    The basic table schema includes a primary key, `glycan_id`, mapping to :attr:`id`.
    Additionally, it includes the mass calculated at the time of serialization under
    the `mass` column, and the entire record structure is pickled and stored under the
    `structure` column.

    The translation to SQL values is carried out by :meth:`.to_sql`, and is restored from
    a query row by :meth:`.from_sql`.
    '''

    #: Default table name used
    __table_name = "GlycanRecord"

    #: The default table schema. Additional
    #: items are added on, replacing /*rest*/
    __table_schema__ = '''
    DROP TABLE IF EXISTS {table_name};
    CREATE TABLE {table_name}(
        glycan_id INTEGER UNIQUE PRIMARY KEY NOT NULL,
        mass float NOT NULL,
        structure TEXT NOT NULL/*rest*/
    );
    '''

    #: The storage for base-class specific
    #: column_data mappings. Add column_data here to
    #: include in all GlycanRecordBase'd objects
    __column_data_map = {}

    @classmethod
    def add_column_data(cls, name, dtype, transform):
        '''
        Function-based approach to modifying the class-specific
        :attr:`.__column_data_map` attribute. Must use :func:`getattr`
        to prevent class-private name mangling

        Parameters
        ----------
        name: str
            Name of the new column_data field
        dtype: str
            The SQL data type to encode the column as
        transform: function
            The function to extract the value of the column_data from a record
        '''
        try:
            getattr(cls, "_{}__column_data_map".format(cls.__name__))[name] = (dtype, transform)
        except:
            setattr(cls, "_{}__column_data_map".format(cls.__name__), {name: (dtype, transform)})

    @classmethod
    def sql_schema(cls, *args, **kwargs):
        '''
        Generate the base table schema for the main GlycanRecord table.

        Parameters
        ----------
        inherits: dict
            Metadata extensions to include in the table schema

        Yields
        ------
        str:
            An SQL script block describing the GlycanRecord table. Descendants
            may yield additional statements.
        '''
        ext_stmts = []
        meta_map = dict(cls.__column_data_map)
        meta_map.update(kwargs.get('inherits', {}))
        for name, type_transformer in meta_map.items():
            template = "{name} {type}\n".format(name=name, type=type_transformer[0])
            ext_stmts.append(template)
        ext_def = ", ".join(ext_stmts)
        if len(ext_def) > 0:
            ext_def = ", " + ext_def
        schema = cls.__table_schema__.format(table_name=cls.table_name)
        schema = schema.replace("/*rest*/", ext_def)
        yield schema

    @classmethod
    def add_index(cls, *args, **kwargs):
        '''
        Generate the base table's indices for fast search

        Yields
        ------
        str:
            The SQL script block describing the mass_index of the GlycanRecord table
        '''
        yield '''CREATE INDEX IF NOT EXISTS mass_index ON {table_name}(mass DESC);'''.format(
            table_name=cls.table_name)
        for index in cls._collect_ext_indices(kwargs.get("inherits", {})):
            yield index

    def __init__(self, structure, motifs=None, dbxref=None, aglycones=None, taxa=None, **kwargs):
        self.structure = structure
        self.motifs = motifs or []
        self.dbxref = dbxref or []
        self.aglycones = aglycones or []
        self.taxa = taxa or []
        self.id = kwargs.get('id')
        self._bound_db = kwargs.get("bound_db")

    @classmethod
    def replicate(cls, record):
        instance = cls(record.structure)
        for name in dir(record):
            if "_" == name[0] or isinstance(getattr(record, name), Callable):
                continue
            try:
                setattr(instance, name, getattr(record, name))
            except:
                pass
        return instance

    def mass(self, average=False, charge=0, mass_data=None, override=None):
        '''
        Calculates the mass of :attr:`structure`. If `override` is not |None|,
        return this instead.

        See Also
        --------
        :meth:`glypy.structure.glycan.Glycan.mass`
        '''
        if override is not None:
            return override
        return self.structure.mass(average=average, charge=charge, mass_data=mass_data)

    def __repr__(self):  # pragma: no cover
        rep = '<{type} {id} {mass}>\n{glycoct}'.format(
                id=(self.id or ''), mass=self.mass(), glycoct=self.structure.serialize("glycoct"),
                type=self.__class__.__name__
            )
        return rep

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop("_bound_db", None)
        return state

    def to_sql(self, id=None, mass_params=None, inherits=None):
        '''
        Translates the :class:`GlycanRecord` instance into SQL.

        Parameters
        ----------
        id: int
            The primary key to use, overwriting :attr:`id` if present. Optional
        mass_params: tuple
            Parameters to pass to :meth:`.mass`. The output is stored
            in the SQL record as the `mass` value
        inherits: dict
            Mapping of inherited column_data properties to include in the record

        Yields
        ------
        str:
            The SQL insert statement adding this record to the database
        '''
        inherits = dict(inherits or {})
        inherits.update(inherits)

        template = '''INSERT INTO {table_name} (glycan_id, mass, structure /*rest*/)
         VALUES ({id}, {mass}, "{structure}" /*values*/);'''
        ext_names = ', '.join(inherits)
        if len(ext_names) > 0:
            ext_names = ', ' + ext_names
        ext_values = ', '.join(["{}".format(v) for k, v in self._collect_ext_data().items()])
        if len(ext_values) > 0:
            ext_values = ', ' + ext_values
        if id is not None:
            self.id = id

        template = template.replace("/*rest*/", ext_names).replace("/*values*/", ext_values)
        values = {}
        values['id'] = self.id
        values['mass'] = self.mass(**(mass_params or {}))
        _bound_db = getattr(self, "_bound_db", None)
        self._bound_db = None
        values['structure'] = pickle.dumps(self)
        self._bound_db = _bound_db
        values['table_name'] = self.__table_name
        yield template.format(**values)

    def to_update_sql(self, mass_params=None, inherits=None, *args, **kwargs):
        '''
        Generates SQL for use with ``UPDATE {table_name} set ... where glycan_id = {id};``.

        Called by :meth:`update`
        '''

        inherits = dict(inherits or {})
        inherits.update(inherits)

        template = '''UPDATE {table_name} SET mass = {mass},
         structure = "{structure}" /*rest*/ WHERE glycan_id = {id};'''

        ext_names = list(inherits)
        ext_values = ["{}".format(v) for k, v in self._collect_ext_data().items()]
        ext_parts = ', '.join(["{} = {}".format(name, value) for name, value in zip(ext_names, ext_values)])
        if len(ext_parts) > 0:
            ext_parts = ", " + ext_parts

        template = template.replace("/*rest*/", ext_parts)
        values = {}
        values['id'] = self.id
        values['mass'] = self.mass(**(mass_params or {}))
        values['structure'] = pickle.dumps(self)
        values['table_name'] = self.__table_name

        yield template.format(**values)

    def update(self, mass_params=None, inherits=None, commit=True, *args, **kwargs):
        """Execute SQL ``UPDATE`` instructions, writing this object's values back to the
        database it was last extracted from.

        Parameters
        ----------
        mass_params: dict
            Parameters passed to :meth:`mass`

        Raises
        ------
        ValueError:
            If the record is not bound to a database

        """
        if self._bound_db is None:
            raise ValueError("Cannot commit an unbound record")
        cur = self._bound_db.cursor()
        for stmt in self.to_update_sql(mass_params=mass_params, inherits=inherits):
            cur.execute(stmt)
        if commit:
            cur.connection.commit()

    @classproperty
    def table_name(cls):
        '''
        A property to get the class-specific table name. Must be used
        to prevent :class:`.RecordDatabase` from mangling references to
        its :attr:`record_type`'s internal table name field on request.
        '''
        return cls.__table_name

    @table_name.setter
    def table_name(cls, value):
        '''
        Setter for :attr:`.table_name`
        '''
        cls.__table_name = value

    @classmethod
    def from_sql(cls, row, *args, **kwargs):
        '''
        Translate a Row object from sqlite3 into a GlycanRecord object

        Parameters
        ----------
        row: sqlite3.Row
            A dict-like object containing the pickled value of the record in
            the `structure` field

        Returns
        -------
        GlycanRecord:
            The unpickled :class:`GlycanRecord` object. Sub-classes may perform
            more complex operations like decompressing or joining other tables in
            the database.
        '''
        record = pickle.loads(bytes(row["structure"]))
        record._bound_db = kwargs.get("database")
        return record

    def _collect_ext_data(self, inherits=None):
        '''
        Apply each column_data mapping transform sequentially, storing each result
        in a |dict| object for generating SQL.
        '''
        meta_map = dict(inherits or {})
        meta_map.update(dict(self.__column_data_map))
        data = {}
        for name, type_transformer in meta_map.items():
            ext_type, transformer = type_transformer
            data[name] = transformer(self)
        return data

    @classmethod
    def _collect_ext_indices(cls, inherits=None):
        meta_map = dict(inherits or {})
        meta_map.update(dict(cls.__column_data_map))
        for name in meta_map:
            yield "CREATE INDEX IF NOT EXISTS {column_name}_{table_name}_index ON {table_name}({column_name});".format(
                column_name=name, table_name=cls.table_name)

    def __eq__(self, other):
        '''
        Equality testing is done between :attr:`structure`
        '''
        try:
            return self.structure == other.structure
        except Exception as e:  # pragma: no cover
            logger.exception(exc_info=e)
            return False

    def __ne__(self, other):  # pragma: no cover
        return not self == other

    def __root__(self):  # pragma: no cover
        return self.structure.root

    def __tree__(self):  # pragma: no cover
        return self.structure


def extract_composition(record, max_size=120):
    '''
    Given a :class:`GlycanRecord`, translate its :attr:`.monosaccharides` property
    into a string suitable for denormalized querying.

    Transforms the resulting, e.g. Counter({u'GlcNA': 6, u'Gal': 4, u'aMan': 2, u'Fuc': 1, u'Man': 1})
    into the string "Gal:4 aMan:2 Fuc:1 GlcNA:6 Man:1" which could be partially matched in
    queries using SQL's LIKE operator.

    Parameters
    ----------
    record: GlycanRecord
        The record to serialize :attr:`.monosaccharides` for
    max_size: int
        The maximum size of the resulting string allowed under the target
        SQL schema

    Returns
    -------
    str:
        The string representation of `record.monosaccharides`.
    '''
    composition_list = ["{}:{}".format(name, count) for name, count in sorted(record.monosaccharides.items())]
    if sum(map(len, composition_list)) + len(composition_list) > max_size:
        raise ValueError(
            "The resulting composition string is larger than {} characters.".format(max_size))
    return '\"' + ' '.join(composition_list) + '\"'


def _query_composition(prefix=None, **kwargs):
    if prefix is None:
        prefix = ''
    else:
        prefix += '.'
    col_name = prefix + "composition"
    composition_list = ["{} LIKE '%{}:{}%'".format(col_name, name, count) for name, count in kwargs.items()]
    return ' AND '.join(composition_list)


def is_n_glycan(record):
    '''
    A predicate testing if the :title:`N-linked Glycan` core motif is present in `record.structure`.

    Returns
    -------
    int:
        0 if |False|, 1 if |True|. Sqlite doesn't have a dedicated |bool| data type.

    See Also
    --------
    :func:`glypy.algorithms.subtree_search.subtree_of`
    '''
    return int(subtree_search.subtree_of(
        n_glycan_core, record.structure, exact=False) == 1)
n_glycan_core = glypy.glycans["N-Linked Core"]


@column_data("composition", "VARCHAR(120)", extract_composition)
@column_data("is_n_glycan", "BOOLEAN", is_n_glycan)
@column_data("glycoct", "TEXT", lambda x: "\"{}\"".format(str(x.structure)))
class GlycanRecord(GlycanRecordBase):
    '''
    An extension of :class:`GlycanRecordBase` to add additional features and better support for extension
    by both metaprogramming and inheritance.
    '''

    __column_data_map = {}

    @querymethod
    def query_like_composition(cls, conn, record=None, prefix=None):
        stmt = "SELECT * FROM {table_name} WHERE " + _query_composition(prefix, **record.monosaccharides) + ";"
        for result in conn.from_sql(conn.execute(stmt)):
            yield result

    @classmethod
    def add_index(cls, *args, **kwargs):
        '''
        Generate the base table's indices for fast search

        Yields
        ------
        str:
            The SQL script block describing the mass_index of the GlycanRecord table
        '''
        for stmt in super(GlycanRecord, cls).add_index(*args, **kwargs):
            yield stmt
        for stmt in cls._collect_ext_indices(kwargs.get("inherits", {})):
            yield stmt

    @property
    def monosaccharides(self):
        '''
        Returns a mapping of the counts of monosaccharides found in :attr:`structure`. Generic names are found
        using :func:`naive_name_monosaccharide`.

        .. note::
            This property is mapped to the the database column ``composition`` by :func:`extract_composition`.


        See Also
        --------
        :func:`extract_composition`
        :func:`naive_name_monosaccharide`
        '''
        res = Counter(map(naive_name_monosaccharide, self.structure))
        res.pop(None, None)
        return res

    @property
    def composition(self):
        return ''.join("{}{}".format(name, number)
                       for name, number in sorted(self.monosaccharides.items()))

    @property
    def is_n_glycan(self):
        '''
        Returns |True| if :attr:`structure` has the :title:`N-linked Glycan` core motif.

        .. note:: This property is mapped to the the database column ``is_n_glycan`` by :func:`is_n_glycan`
        '''
        return bool(is_n_glycan(self))

    @classmethod
    def sql_schema(cls, *args, **kwargs):
        meta_map = dict(cls.__column_data_map)
        meta_map.update(kwargs.pop("inherits", {}))
        return super(GlycanRecord, cls).sql_schema(inherits=_resolve_column_data_mro(cls))

    def to_sql(self, *args, **kwargs):
        inherits = _resolve_column_data_mro(self.__class__)
        return super(GlycanRecord, self).to_sql(*args, inherits=inherits, **kwargs)

    def to_update_sql(self, *args, **kwargs):
        inherits = _resolve_column_data_mro(self.__class__)
        inherits = inherits or _resolve_column_data_mro(self.__class__)
        kwargs['inherits'] = inherits
        return super(GlycanRecord, self).to_update_sql(*args, **kwargs)

    def update(self, mass_params=None, inherits=None, commit=True, *args, **kwargs):
        inherits = inherits or _resolve_column_data_mro(self.__class__)
        kwargs['inherits'] = inherits
        super(GlycanRecord, self).update(mass_params=mass_params, *args, **kwargs)

    def _collect_ext_data(self):
        inherits = _resolve_column_data_mro(self.__class__)
        data = super(GlycanRecord, self)._collect_ext_data(inherits=inherits)
        return data

    @classmethod
    def _collect_ext_indices(cls, inherits=None):
        inherits = _resolve_column_data_mro(cls)
        data = super(GlycanRecord, cls)._collect_ext_indices(inherits=inherits)
        return data


class GlycanRecordWithTaxon(GlycanRecord):
    __taxa_table_schema__ = '''
    DROP TABLE IF EXISTS RecordTaxonomy;
    CREATE TABLE RecordTaxonomy(
        glycan_id INTEGER NOT NULL,
        taxon_id INTEGER NOT NULL
    );
    '''

    @classmethod
    def sql_schema(cls, *args, **kwargs):
        for line in super(GlycanRecordWithTaxon, cls).sql_schema(*args, **kwargs):
            yield line
        yield cls.__taxa_table_schema__

    @classmethod
    def add_index(cls, *args, **kwargs):
        for line in super(GlycanRecordWithTaxon, cls).add_index(*args, **kwargs):
            yield line
        yield "CREATE INDEX IF NOT EXISTS TaxonomyIndex ON RecordTaxonomy(taxon_id);"
        yield "CREATE INDEX IF NOT EXISTS TaxonomyIndex2 ON RecordTaxonomy(glycan_id);"

    def to_sql(self, *args, **kwargs):
        for line in super(GlycanRecordWithTaxon, self).to_sql(*args, **kwargs):
            yield line
        for taxon in self.taxa:
            yield "INSERT OR REPLACE INTO RecordTaxonomy (glycan_id, taxon_id) VALUES ({}, {});".format(
                self.id, int(taxon.tax_id))

    @querymethod
    def query_by_taxon_id(cls, conn, taxon_ids):
        # Passed an iterable of taxa to search
        try:
            taxon_ids = tuple(taxon_ids)
            taxon_ids_token = str(taxon_ids)
            if len(taxon_ids) == 1:
                taxon_ids_token = taxon_ids_token.replace(',', '')

            return conn.from_sql(conn.execute(
                """SELECT {table_name}.* FROM {table_name} JOIN RecordTaxonomy taxa ON
                   taxa.glycan_id = {table_name}.glycan_id WHERE taxa.taxon_id IN [!SUB!];
                """.replace('[!SUB!]', taxon_ids_token)))
        # Passed a single taxon
        except:
            return conn.from_sql(conn.execute(
                """SELECT {table_name}.* FROM {table_name} JOIN RecordTaxonomy taxa ON
                   taxa.glycan_id = {table_name}.glycan_id WHERE taxa.taxon_id = ?;""",
                (int(taxon_ids),)))


class RecordDatabase(object):
    '''
    A wrapper around an Sqlite3 database for storing and searching GlycanRecord
    objects.

    This class defines a handful general data access methods as well as the ability
    to directly write SQL queries against the database.

    Calls :meth:`.apply_schema`. If ``records`` is not |None|, calls :meth:`.apply_indices`.

    ``record_type``, the stored class type is used for inferring the table schema,
    table name, and :meth:`GlycanRecord.from_sql` function.

    If ``records`` is not provided, no records are added. If records are provided, they are inserted
    with :meth:`.load_data` and afterwards :meth:`.apply_indices` is called.

    Attributes
    ----------
    connection_string: |str|
        The path to the Sqlite database file, or the ":memory:" special
        keyword defining the database to be held directly in memory.
    record_type: :class:`type`
        The class type of the records assumed to be stored in this database. The stored
        class type is used for inferring the table schema, table name, and :meth:`GlycanRecord.from_sql`
        function.

    Parameters
    ----------
    connection_string: str
        The path to the Sqlite database file, or the ":memory:" special
        keyword defining the database to be held directly in memory.
    record_type: type
        The class type of the records assumed to be stored in this database. Defaults to :class:`GlycanRecord`
    records: list
        A list of `record_type` records to insert immediately on table creation.
    '''
    def __init__(self, connection_string=":memory:", record_type=GlycanRecord, records=None, flag='c'):

        created_new = False
        if connection_string == ":memory:" or not os.path.exists(connection_string):
            created_new = True

        if flag == "c":
            pass
            # If 'c', create if not present, but don't clear old data
        elif flag == "w":
            created_new = True
            # If 'w', clear the table before taking any operations

        self.connection_string = connection_string
        self.connection = sqlite3.connect(connection_string)
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor

        # Check to see if the record type matches what is already
        # stored in the database.
        try:
            if record_type is None:
                self.record_type = GlycanRecord
            else:
                self.record_type = record_type
            rec = next(iter(self))
            if not type(rec) == record_type:
                if record_type is not None:
                    logger.warn("Record class {0} is not the same as {1}. Using {0}".format(type(rec), record_type))
                    print("Record class {0} is not the same as {1}. Using {0}".format(type(rec), record_type))
                record_type = type(rec)
        except (StopIteration, sqlite3.OperationalError):
            # If record_type is None and no types were inferred, use GlycanRecord
            if record_type is None:
                record_type = GlycanRecord

        self.record_type = record_type
        self._id = 0

        if records is not None:
            self.apply_schema()
            self.load_data(records)
            self.apply_indices()
        elif created_new:
            self.apply_schema()
        else:
            self._id = len(self)
        try:
            self._patch_querymethods()
        except Exception as e:
            logger.error(exc_info=e)

    def apply_schema(self):
        '''
        Executes each SQL block yielded by :attr:`.record_type`'s :meth:`.sql_schema` class method.
        Commits all pending changes.

        Called during initialization if the database is newly created.

        .. danger::
            The SQL table definition statements generated may drop existing tables. Calling
            this function on an already populated table can cause data loss.
        '''
        self.executescript('''
        DROP TABLE IF EXISTS metadata;
        CREATE TABLE metadata(
            name VARCHAR(40) PRIMARY KEY,
            content TEXT
        );''')
        self.executescript('\n'.join(self.record_type.sql_schema()))
        self.commit()
        self._id = 0

    def apply_indices(self):
        '''
        Executes each SQL block yielded by :attr:`.record_type`'s :meth:`.add_index` class method.
        Commits all pending changes.

        May be called during initialization if data was added.
        '''
        for ix_stmt in self.record_type.add_index():

            self.execute(ix_stmt)
        self.commit()

    def load_data(self, record_list, commit=True, set_id=True, cast=True, **kwargs):
        '''
        Given an iterable of :attr:`.record_type` objects,
        assign each a primary key value and insert them into the
        database.

        Forwards all ``**kwargs`` to :meth:`to_sql` calls.

        Parameters
        ----------
        record_list: GlycanRecord or iterable of GlycanRecords
        commit: bool
            Whether or not to commit all changes to the database
        set_id: bool
        cast: bool
        '''
        if not isinstance(record_list, Iterable):
            record_list = [record_list]
        for record in record_list:
            if set_id:
                self._id += 1
                record.id = self._id
            if cast and not isinstance(record, self.record_type):
                record = self.record_type.replicate(record)
            for stmt in record.to_sql(**kwargs):
                try:
                    self.connection.execute(stmt)
                except:
                    print(stmt)
                    raise
        if commit:
            self.commit()

    def get_metadata(self, key=None):
        """Retrieve a value from the key-value store
        in the database's metadata table.

        If `key` is |None| all of the keys will be retrieved
        and returned as a |dict|.

        Parameters
        ----------
        key : str, optional
            The key value to retrieve.

        Returns
        -------
        any or |dict|
        """
        if key is None:
            return {
                row[0]: pickle.loads(bytes(row[1])) for row in self.execute("SELECT name, content FROM metadata")
            }
        res = [pickle.loads(bytes(row[0])) for row in self.execute("SELECT content FROM metadata WHERE name = ?", (key,))]
        if len(res) > 0:
            return res[0]
        else:
            raise KeyError("Key {} not found".format(key))

    def set_metadata(self, key, value):
        """Set a key-value pair in the database's metadata table

        Parameters
        ----------
        key : str
            Key naming value
        value : any
            Value to store in the metadata table
        """
        self.execute("INSERT OR REPLACE INTO metadata (name, content) VALUES (?, ?);", (key, pickle.dumps(value)))
        self.commit()

    def bind(self, record):
        record._bound_db = self

    def _patch_querymethods(self):
        """Patch on additional methods defined on :attr:`record_type`
        """
        for name, value in _resolve_querymethods_mro(self.record_type).items():
            if isinstance(value, QueryMethod):
                setattr(self, name, value.bind(self))

    def __len__(self):
        """The number of records in the database

        Returns
        -------
        int
        """
        res = (self.execute("SELECT count(glycan_id) FROM {table_name};").next())["count(glycan_id)"]
        return res or 0

    def create(self, structure, *args, **kwargs):
        '''
        A convenience function for creating a database entry for |Glycan| `structure`. Passes
        along all arguments to :attr:`record_type` initialization methods.

        Parameters
        ----------
        structure: :class:`Glycan`
        commit: bool
            If |True|, commit changes after adding this record. Defaults to |True|.
        '''
        commit = kwargs.pop("commit", True)
        record = self.record_type(structure=structure, *args, **kwargs)
        self.load_data([record], commit=commit)
        return record

    def __getitem__(self, keys):
        '''
        Look up records in the database by primary key. Also accepts :class:`slice` objects.

        Returns
        -------
        :class:`.record_type` or :class:`list` of :class:`.record_type`
        '''
        results = []
        key_type = int
        if isinstance(keys, slice):
            key_type = slice
        if isinstance(keys, (tuple, list, set)):
            key_type = tuple
        else:
            keys = int(keys)
        if key_type is int:
            results = list(self.from_sql(
                self.execute("SELECT * FROM {table_name} WHERE glycan_id = ?", (keys,))))
        elif key_type is slice:
            begin = keys.start or 1
            end = keys.stop or len(self)

            results = list(self.from_sql(
                self.execute(
                    "SELECT * FROM {table_name} WHERE glycan_id BETWEEN ? AND ?", begin, end)))
        elif key_type is tuple:
            group = tuple(map(int, keys))
            if len(group) == 1:
                group = str(group).replace(',', '')
            results = list(self.from_sql(
                self.execute(
                    "SELECT * FROM {table_name} WHERE glycan_id IN {}".format(
                        group, table_name=self.record_type.table_name))))

        if len(results) == 0 and key_type is int:
            raise IndexError("No record found for %r" % keys)
        list(map(self.bind, results))
        if len(results) > 1:
            return results
        elif key_type is int:
            return results[0]
        else:
            return results

    def __iter__(self):
        '''
        Iterate sequentially over each entry in the database.
        '''
        for row in self.from_sql(self.execute(self.stn("select * from {table_name};"))):
            yield row

    def __contains__(self, key):
        try:
            self[int(key)]
            return True
        except IndexError:
            return False

    def __repr__(self):  # pragma: no cover
        rep = "<RecordDatabase {} records>".format(len(self))
        return rep

    def sub_table_name(self, string, key='table_name'):
        '''
        A convenience function called to substitute in the primary table name
        into raw SQL queries. By default it looks for the token {table_name}.
        '''
        return string.format(**{key: self.record_type.table_name})

    stn = sub_table_name

    def execute(self, query, *args, **kwargs):
        '''
        A wrapper around :meth:`sqlite3.Connection.execute`. Will format
        the query string to substitute in the main table name if the {table_name}
        token is present
        '''
        return self.connection.execute(self.stn(query), *args, **kwargs)

    def executemany(self, query, param_iter, *args, **kwargs):
        '''
        A wrapper around :meth:`sqlite3.Connection.executemany`. Will format
        the query string to substitute in the main table name if the {table_name}
        token is present.
        '''
        return self.connection.executemany(self.stn(query), param_iter, *args, **kwargs)

    def executescript(self, script, *args, **kwargs):
        '''
        A wrapper around :meth:`sqlite3.Connection.executescript`.
        '''
        return self.connection.executescript(script, *args, **kwargs)

    def commit(self):
        '''
        A wrapper around :meth:`sqlite3.Connection.commit`. Writes pending changes to the database.
        '''
        self.connection.commit()

    def rollback(self):
        '''
        A wrapper around :meth:`sqlite3.Connection.rollback`. Reverses the last set of changes to the database.
        '''
        self.connection.rollback()

    def close(self):
        self.connection.close()

    def _find_boundaries(self, mass, tolerance):
        spread = mass * tolerance
        return (mass - spread, mass + spread)

    def ppm_match_tolerance_search(self, mass, tolerance, mass_shift=0):
        '''
        Rapidly search the database for entries with a recorded mass within
        ``tolerance`` parts per million mass error of ``mass``.

        :math:`[mass - (tolerance * mass), mass + (tolerance * mass)]`

        '''
        lower, upper = self._find_boundaries(mass + mass_shift, tolerance)
        results = self.execute("SELECT * FROM {table_name}\
         WHERE mass BETWEEN ? AND ?;", (lower, upper))
        for result in results:
            yield self.record_type.from_sql(result, database=self)

    def from_sql(self, rows, from_sql_fn=None):
        """Convenience function to convert `rows` into objects through `from_sql_fn`,
        by default, :meth:`self.record_type.from_sql`

        Parameters
        ----------
        rows : sqlite3.Row or an iterable of sqlite3.Row
            Collection of objects to convert
        from_sql_fn : function, optional
            Function to perform the conversion. Defaults to :meth:`self.record_type.from_sql`

        Yields
        -------
        Type returned by `from_sql_fn`
        """
        if from_sql_fn is None:
            from_sql_fn = self.record_type.from_sql
        if isinstance(rows, sqlite3.Row):
            rows = [rows]
        for row in rows:
            record = from_sql_fn(row, self)
            self.bind(record)
            yield record

#: Open a database
dbopen = RecordDatabase




from matplotlib.textpath import TextPath
import matplotlib.patches as patches

from glypy.io import iupac

from .common import MonosaccharidePatch
from .geometry import centroid, Affine2D
from .symbolic_nomenclature import SymbolicNomenclatureBase

default_line_color = 'lightgrey'
# default_line_color = 'black'
default_scale_factor = 1.0
zorder = 2
line_weight = 0.5


class IUPACTextSymbolicNomenclature(SymbolicNomenclatureBase):
    default_line_color = 'lightgrey'

    def draw_text(self, ax, x, y, text, scale=0.1, **kwargs):
        fs = kwargs.get("fontsize", 2) * scale * 0.13
        t_path = TextPath((x - 0.1, y), s=text, size=fs)
        center = centroid(t_path)
        dist = -1 * (center - (x, y))
        t_path_shifted = t_path.transformed(Affine2D().translate(*dist))
        patch = patches.PathPatch(t_path_shifted, facecolor="black", lw=line_weight / 20., zorder=4)
        a = ax.add_patch(patch)
        return a

    def draw(self, monosaccharide, x, y, ax, tree_node=None, scale=0.1, **kwargs):
        try:
            text = iupac.to_iupac(monosaccharide)
        except Exception:
            text = monosaccharide
        if text[0] == 'a':
            text = text.rsplit("-", 1)[1]
        elif text[0] == 'b':
            text = text.rsplit("-", 1)[1]
        node = self.draw_text(ax, x, y, text, scale=scale * 3.)
        return MonosaccharidePatch(saccharide_label=(node,))



cimport cython

from cpython.object cimport PyObject
from cpython.ref cimport Py_INCREF
from cpython.list cimport PyList_Append
from cpython.dict cimport PyDict_GetItem, PyDict_Size, PyDict_Next, PyDict_SetItem
from cpython.int cimport PyInt_AsLong, PyInt_FromLong


from libc.string cimport strcmp, memcpy, strlen, strncpy, strcpy
from libc.stdlib cimport malloc, free, realloc, atoi, calloc

from glypy._c.compat cimport (
    PyStr_FromStringAndSize, PyStr_AsUTF8AndSize, PyInt_FromString,
    PyStr_InternInPlace)

cdef extern from * nogil:
    int printf (const char *template, ...)
    void qsort (void *base, unsigned short n, unsigned short w, int (*cmp_func)(void*, void*))


# Parser Helpers


cpdef tuple parse_name_count(str string):
    cdef:
        size_t i
        Py_ssize_t n
        Py_ssize_t boundary
        char* cstring
        str elt
        object count

    cstring = PyStr_AsUTF8AndSize(<str>string, &n)
    elt = None
    count = 0
    for i in range(n):
        if cstring[i] == ':' and i < n:
            elt = PyStr_FromStringAndSize(cstring, i)
            count = PyInt_FromString(&cstring[i + 1], NULL, 10)
            return elt, count
    raise ValueError()


cpdef tuple get_parse_tokens(object string):
    cdef:
        size_t i
        Py_ssize_t n
        Py_ssize_t elt_start, elt_end
        list tokens
        str reduced, token
        char* cstring
    if not isinstance(string, str):
        string = str(string)
    reduced = None
    tokens = []
    cstring = PyStr_AsUTF8AndSize(<str>string, &n)
    elt_start = 1
    elt_end = -1
    if n < 2 or (n > 2 and cstring[0] != '{'):
        raise ValueError("Malformed glycan composition!")
    if n == 2:
        if cstring[0] == '{' and cstring[1] == '}':
            return [], reduced
        else:
            raise ValueError("Malformed glycan composition!")
    for i in range(1, n - 1):
        if (cstring[i] == ';' and cstring[i + 1] == ' '):
            elt_end = i
            token = PyStr_FromStringAndSize(
                &cstring[elt_start], elt_end - elt_start)
            PyList_Append(tokens, token)
            elt_start = i + 2
        elif cstring[i] == '}':
            elt_end = i
            token = PyStr_FromStringAndSize(
                &cstring[elt_start], elt_end - elt_start)
            PyList_Append(tokens, token)
            elt_start = i + 1
        elif cstring[i] == '$':
            reduced = PyStr_FromStringAndSize(&cstring[i + 1], n - (i + 1))
            break
    else:
        if cstring[i + 1] == '}':
            elt_end = i + 1
            token = PyStr_FromStringAndSize(
                &cstring[elt_start], elt_end - elt_start)
            PyList_Append(tokens, token)
    return tokens, reduced

cdef struct string_cell:
    char* string
    size_t size


# Serialization Helpers

DEF ARRAY_SIZE = 2 ** 12
cdef string_cell[ARRAY_SIZE] str_ints
cdef dict str_int_overflow_intern_cache = dict()

cdef void init_str_ints():
    cdef int i
    cdef Py_ssize_t z
    cdef str pa
    cdef char* a
    cdef char* atemp

    for i in range(ARRAY_SIZE):
        pa = str(i)
        atemp = PyStr_AsUTF8AndSize(pa, &z)
        a = <char*>malloc(sizeof(char) * z)
        strcpy(a, atemp)
        a[z] = "\0"
        str_ints[i].string = a
        str_ints[i].size = z


init_str_ints()



cdef int get_str_int(int i, string_cell* out):
    '''Get a `string_cell` for a given integer.

    For integers < 2 ** 12, go through the `str_ints` array
    fast path, otherwise use `str_int_overflow_intern_cache`.

    `str_int_overflow_intern_cache` keeps a mapping from PyInt to
    an immortalized PyStr whose internal buffer is used to serve
    the `char*` for `string_cell` instances larger than  2 ** 12 - 1.
    '''
    cdef:
        PyObject* ptemp
        object py_i
        str pa
        Py_ssize_t z

    if i < ARRAY_SIZE and i >= 0:
        out[0] = str_ints[i]
        return 0

    py_i = PyInt_FromLong(i)
    ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
    if ptemp == NULL:
        pa = str(py_i)
        Py_INCREF(pa)
        PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)
    else:
        pa = <str>ptemp
    atemp = PyStr_AsUTF8AndSize(pa, &z)
    out.string = atemp
    out.size = z
    return 0


cdef struct monosaccharide_count_cell:
    char* string
    Py_ssize_t size
    int count
    float mass


cdef int cmp_monosaccharide_count_cell(const void* lhs, const void* rhs) nogil:
    cdef monosaccharide_count_cell* left = <monosaccharide_count_cell*>lhs
    cdef monosaccharide_count_cell* right = <monosaccharide_count_cell*>rhs

    if left.mass < right.mass:
        return -1
    if right.mass < left.mass:
        return 1
    if left.string == NULL:
        if right.string == NULL:
            return 0
        else:
            return -1
    if right.string == NULL:
        return 1
    return strcmp(left.string, right.string)


cdef dict intern_table = dict()


cpdef str _prepare_glycan_composition_string(dict composition):
    cdef size_t n = PyDict_Size(composition)
    cdef monosaccharide_count_cell* cells = <monosaccharide_count_cell*>malloc(sizeof(monosaccharide_count_cell) * n)
    cdef:
        Py_ssize_t i
        size_t j, z, k
        PyObject* pkey
        PyObject* pvalue
        PyObject* ptmp
        str name
        int value
        char* text_buffer
        string_cell int_conv

    i = 0
    j = 0
    z = 2
    while PyDict_Next(composition, &i, &pkey, &pvalue):
        value = PyInt_AsLong(<object>pvalue)
        if value == 0:
            cells[j].string = NULL
            cells[j].size = cells[j].count = 0
            cells[j].mass = -1
        else:
            name = str(<object>pkey)
            ptmp = PyDict_GetItem(intern_table, name)
            if ptmp == NULL:
                Py_INCREF(name)
                PyDict_SetItem(intern_table, name, name)
            else:
                name = <str>ptmp

            cells[j].string = PyStr_AsUTF8AndSize(<str>name, &cells[j].size)
            # The separators, of which there are two extra from the last monosaccharide
            z += 3
            z += cells[j].size
            # Number of digits
            z += (value // 10) + 1
            cells[j].count = value
            cells[j].mass = (<object>pkey).mass()
        j += 1
    qsort(<void*>cells, j, sizeof(monosaccharide_count_cell), cmp_monosaccharide_count_cell)
    if z == 2:
        free(cells)
        return "{}"
    text_buffer = <char*>malloc(sizeof(char) * z)

    k = 0
    text_buffer[k] = '{'
    k += 1
    for j in range(n):
        if cells[j].string == NULL:
            continue
        memcpy(&text_buffer[k], cells[j].string, cells[j].size)
        k += cells[j].size
        text_buffer[k] = ':'
        k += 1
        value = cells[j].count
        get_str_int(value, &int_conv)
        memcpy(&text_buffer[k], int_conv.string, int_conv.size)
        k += int_conv.size
        text_buffer[k] = ';'
        k += 1
        text_buffer[k] = ' '
        k += 1
    # We've written the "; " separator for the last monosaccharide,
    # so now we must rewind two steps and write the terminal '}' over
    # them and shrink k by 1 to reflect actual length of the string.
    if k > 1:
        text_buffer[k - 2] = '}'
        k -= 1
    else:
        text_buffer[k] = '}'
        k += 1
    free(cells)
    result = PyStr_FromStringAndSize(text_buffer, k)
    return result



def reduced_end_compat(reduced_end):
    if reduced_end is None:
        return
    if not hasattr(reduced_end, "base_composition"):
        if len(reduced_end.links) == 0:
            reduced_end.base_composition = reduced_end.composition.clone()
        else:
            reduced_end.base_composition = reduced_end.composition.clone()
            for pos, link in reduced_end.links.items():
                reduced_end.base_composition += link.parent_loss
    if reduced_end.valence != 1:
        try:
            reduced_end.drop_substituent(2)
        except IndexError:
            pass
        reduced_end.valence = 1


#: Copies of the functions of the same name instrumented for determining where exactly an
#: equality search failed.
def exact_ordering_equality(self, other, substituents=True, visited=None):
    '''
    Performs equality testing between two monosaccharides where
    the exact position (and ordering by sort) of links must to match between
    the input |Monosaccharide| objects

    Returns
    -------
    |bool|
    '''
    if visited is None:
        visited = set()
    if (self.id, other.id) in visited:
        return True

    visited.add((self.id, other.id))

    if self._flat_equality(other):
        if substituents:
            for a_sub, b_sub in zip(self.substituents(), other.substituents()):
                if a_sub != b_sub:
                    return False
        for a_mod, b_mod in zip(self.modifications.items(), other.modifications.items()):
            if a_mod != b_mod:
                return False
        for a_child, b_child in zip(self.children(), other.children()):
            if a_child[0] != b_child[0]:
                return False
            if not a_child[1].exact_ordering_equality(b_child[1],
                                                      substituents=substituents,
                                                      visited=visited):
                return False
        return True
    return False


def topological_equality(self, other, substituents=True, visited=None):
    '''
    Performs equality testing between two monosaccharides where
    the exact ordering of child links does not have to match between
    the input |Monosaccharide|s, so long as an exact match of the
    subtrees is found

    Returns
    -------
    |bool|
    '''
    if visited is None:
        visited = set()
    if (self.id, other.id) in visited:
        print ("Already visited ", (self.id, other.id))
        return True
    if self._flat_equality(other) and (not substituents or self._match_substituents(other)):
        taken_b = set()
        b_children = list(other.children())
        a_children = list(self.children())
        for a_pos, a_child in a_children:
            matched = False
            for b_pos, b_child in b_children:
                if (b_pos, b_child.id) in taken_b:
                    continue
                if a_child.topological_equality(b_child,
                                                substituents=substituents,
                                                visited=visited):
                    matched = True
                    taken_b.add((b_pos, b_child.id))
                    break
            if not matched and len(a_children) > 0:
                print ("not matched and len(a_children) > 0", self.id, other.id)
                return False
        if len(taken_b) != len(b_children):
            print("len(taken_b) != len(b_children)", self.id, other.id)
            return False
        return True
    print("Not Flat-Equal", self.id, other.id)
    return False


def _match_substituents(self, other):
    '''
    Helper method for matching substituents in an order-independent
    fashion. Used by :meth:`topological_equality`
    '''
    taken_b = set()
    b_num_unknown = len(other.substituent_links[-1])
    unknown_cntr = 0
    b_substituents = list(other.substituents())
    cntr = 0
    for a_pos, a_substituent in self.substituents():
        matched = False
        cntr += 1
        for b_pos, b_substituent in b_substituents:
            if b_pos in taken_b:
                if b_pos != -1:
                    continue
                else:
                    if unknown_cntr < b_num_unknown:
                        unknown_cntr += 1
                    else:
                        continue
            if b_substituent == a_substituent:
                matched = True
                taken_b.add(b_pos)
                break
        if not matched and cntr > 0:
            return False
    if len(taken_b) + unknown_cntr != len(b_substituents):
        return False
    return True


def reparse_database(database):
    from glypy.io import glycoct
    for record in database:
        ct_str = str(record.structure)
        structure = glycoct.loads(ct_str)
        assert structure.mass() == record.structure.mass()
        record.structure = structure
        record.update()



import unittest
from glypy.composition import composition_transform
from glypy.algorithms import database
from .common import load


# class GlycanRecordTest(unittest.TestCase):

#     def test_record_creation(self):
#         rec = database.GlycanRecord(load("broad_n_glycan"))

#         self.assertEqual(rec.structure.mass(), rec.mass())

#         saccharides = {'Hex': 7, u'HexNAc': 6, 'dHex': 1}
#         self.assertEqual(saccharides, rec.monosaccharides)

#         self.assertEqual(database.extract_composition(rec), '"Hex:7 HexNAc:6 dHex:1"')

#     def test_to_sql(self):
#         rec = database.GlycanRecord(load("broad_n_glycan"))
#         db = database.RecordDatabase()
#         db.apply_schema()
#         db.create(load("broad_n_glycan"))
#         self.assertEqual(db[1], rec)

#     def test_update(self):
#         db = database.RecordDatabase()
#         db.apply_schema()
#         db.create(load("broad_n_glycan"))
#         rec = db[1]
#         composition_transform.derivatize(rec.structure, "methyl")
#         rec.update()
#         dup = db[1]
#         self.assertAlmostEqual(rec.mass(), dup.mass(), 3)

#     def test_replicate(self):
#         db = database.RecordDatabase()
#         db.apply_schema()
#         db.create(load("broad_n_glycan"))
#         rec = db[1]
#         self.assertEqual(db.record_type.replicate(rec), rec)


# class RecordDatabaseTest(unittest.TestCase):

#     def test_load_data(self):
#         rec = database.GlycanRecord(load("broad_n_glycan"))
#         rec2 = database.GlycanRecord(load("complex_glycan"))
#         db = database.RecordDatabase(records=[rec, rec2])
#         self.assertTrue(db[1] == rec)
#         self.assertTrue(db[2] == rec2)

#     def test_ppm_search(self):
#         rec = database.GlycanRecord(load("broad_n_glycan"))
#         rec2 = database.GlycanRecord(load("complex_glycan"))
#         db = database.RecordDatabase(records=[rec, rec2])
#         self.assertEqual(rec, (db.ppm_match_tolerance_search(rec.mass(), 1e-5)).next())

#     def test_record_type_inference(self):
#         db = database.dbopen("./test_data/test.db", database.GlycanRecordWithTaxon, flag='w')
#         db.create(load("complex_glycan"))
#         db.commit()
#         self.assertEqual(db.record_type, database.GlycanRecordWithTaxon)
#         db.close()
#         db = database.dbopen("./test_data/test.db")
#         self.assertEqual(db.record_type, database.GlycanRecordWithTaxon)
#         db.close()
#         db = database.dbopen("./test_data/test.db", record_type=None)
#         self.assertEqual(db.record_type, database.GlycanRecordWithTaxon)

#     def test_metadata(self):
#         db = database.dbopen()
#         db.set_metadata("Spam", {"Ham", "Eggs"})
#         self.assertEqual(db.get_metadata("Spam"), {"Ham", "Eggs"})
#         self.assertEqual(len(db), 0)

if __name__ == '__main__':
    unittest.main()



import unittest
import hjson

from glypy.structure import named_structures, constants, monosaccharide, substituent, glycan
from glypy.composition import structure_composition, Composition, composition_transform
from glypy.io import glycoct

from .common import StringIO, load, pickle


Monosaccharide = monosaccharide.Monosaccharide


monosaccharide_structures = hjson.load(
    open("./glypy/structure/data/monosaccharides.hjson"))

wiki_masses = {
    "Iduronic Acid": 194.04,
    "Bacillosamine": 162.10,
    "Allose": 180.06,
    "Neu5Ac": 309.105,
    "Fru": 180.06,
}

ReducedEnd = monosaccharide.ReducedEnd


class MonosaccharideTests(unittest.TestCase):
    _file_path = "./test_data/glycoct.txt"
    glycan = next(iter(glycoct.read(_file_path)))

    def test_depth(self):
        self.assertEqual(monosaccharide.depth(load("complex_glycan").root), 7)

    def test_from_glycoct(self):
        s = self.glycan.root.serialize('glycoct')
        b = StringIO(s)
        g = next(iter(glycoct.read(b)))
        self.assertEqual(g.root.serialize('glycoct'), s)

    def test_named_structure_masses(self):
        for name, mass in wiki_masses.items():
            structure = named_structures.monosaccharides[name]
            self.assertAlmostEqual(mass, structure.mass(), 2)

    def test_ring_limit_modification(self):
        structure = named_structures.monosaccharides['Hex']
        self.assertRaises(
            IndexError, lambda: structure.add_modification('d', 8))

    def test_occupancy_limit_modification(self):
        structure = named_structures.monosaccharides['Hex']
        structure.add_modification('d', 4)
        self.assertRaises(
            ValueError, lambda: structure.add_modification('d', 4))

    def test_ring_limit_substituent(self):
        structure = named_structures.monosaccharides['Hex']
        self.assertRaises(
            IndexError, lambda: structure.add_substituent('methyl', 8))

    def test_occupancy_limit_substituent(self):
        structure = named_structures.monosaccharides['Hex']
        structure.add_substituent('methyl', 4)
        self.assertRaises(
            ValueError, lambda: structure.add_substituent('methyl', 4))

    def test_add_remove_modifcations(self):
        for name, mass in wiki_masses.items():
            structure = named_structures.monosaccharides[name]
            ref = structure.clone()
            self.assertAlmostEqual(ref.mass(), structure.mass())
            open_sites, unknowns = structure.open_attachment_sites()
            n_sites = len(open_sites)
            comp_delta = n_sites * \
                structure_composition.modification_compositions[
                    constants.Modification.d]()
            for site in open_sites:
                structure.add_modification(constants.Modification.d, site)
            self.assertEqual(
                structure.total_composition(), ref.total_composition() + comp_delta)
            self.assertEqual([], structure.open_attachment_sites()[0])
            for site in open_sites:
                structure.drop_modification(site, constants.Modification.d)
            self.assertEqual(
                structure.total_composition(), ref.total_composition())

    def test_add_remove_substituents(self):
        for name, mass in wiki_masses.items():
            structure = named_structures.monosaccharides[name]
            ref = structure.clone()
            self.assertAlmostEqual(ref.mass(), structure.mass())
            open_sites, unknowns = structure.open_attachment_sites()
            n_sites = len(open_sites)
            mass_delta = substituent.Substituent(
                'methyl').mass() * n_sites - Composition("H2").mass * n_sites
            ping = True
            for site in open_sites:
                if ping:
                    structure.add_substituent(
                        substituent.Substituent('methyl'), position=site)
                    ping = False
                else:
                    structure.add_substituent('methyl', position=site)
                    ping = True
            self.assertAlmostEqual(structure.mass(), ref.mass() + mass_delta)
            for site in open_sites:
                structure.drop_substituent(
                    site, substituent.Substituent('methyl'))
            self.assertAlmostEqual(structure.mass(), ref.mass())

    def test_validate_drop_substituents(self):
        structure = named_structures.monosaccharides['Hex']
        self.assertRaises(IndexError, lambda: structure.drop_substituent(99))
        self.assertRaises(IndexError, lambda: structure.drop_substituent(4))
        self.assertRaises(IndexError, lambda: structure.add_substituent(
            "n-acetyl", 3, max_occupancy=4).add_substituent(
            "methyl", 3, max_occupancy=4).drop_substituent(3, "n-glycolyl"))

    def test_add_remove_monosaccharides(self):
        for name, mass in wiki_masses.items():
            structure = named_structures.monosaccharides[name]
            ref = structure.clone()
            self.assertAlmostEqual(ref.mass(), glycan.Glycan(structure).mass())
            open_sites, unknowns = structure.open_attachment_sites()
            n_sites = len(open_sites)
            mass_delta = named_structures.monosaccharides[
                "Hex"].mass() * n_sites - Composition("H2O").mass * n_sites
            for site in open_sites:
                structure.add_monosaccharide(
                    named_structures.monosaccharides["Hex"], position=site, child_position=3)
            self.assertAlmostEqual(
                glycan.Glycan(structure).mass(), ref.mass() + mass_delta)
            for site in open_sites:
                structure.drop_monosaccharide(site)
            self.assertAlmostEqual(glycan.Glycan(structure).mass(), ref.mass())

    def test_validate_drop_monosacharide(self):
        structure = named_structures.monosaccharides['Hex']
        self.assertRaises(
            IndexError, lambda: structure.drop_monosaccharide(99))
        self.assertRaises(ValueError, lambda: structure.drop_monosaccharide(4))
        self.assertRaises(ValueError, lambda: structure.add_monosaccharide(
            named_structures.monosaccharides["Hex"], 3, max_occupancy=4).add_monosaccharide(
            named_structures.monosaccharides["Hex"], 3, max_occupancy=4).drop_monosaccharide(3))

    def test_validate_enums(self):
        structure = named_structures.monosaccharides['Hex']

        def t():
            structure.anomer = 5
        self.assertRaises(KeyError, t)

        def t():
            structure.stem = "gibberish"
        self.assertRaises(KeyError, t)

        def t():
            structure.superclass = "monose"
        self.assertRaises(KeyError, t)

        def t():
            structure.configuration = 'not-real'
        self.assertRaises(KeyError, t)

    def test_validate_reducing_end(self):
        structure = named_structures.monosaccharides['Hex']
        composition = structure.total_composition()
        structure.reducing_end = ReducedEnd()
        self.assertEqual(structure.total_composition(), composition + Composition("H2"))
        structure.reducing_end = True
        self.assertEqual(structure.total_composition(), composition + Composition("H2"))
        self.assertEqual(structure.total_composition(), structure.clone().total_composition())
        self.assertEqual(structure.total_composition(), pickle.loads(pickle.dumps(structure)).total_composition())
        structure.reducing_end = None
        self.assertEqual(structure.total_composition(), composition)
        structure.reducing_end = True
        composition_transform.derivatize(structure, "methyl")
        self.assertEqual(structure.mass(), structure.clone().mass())

    def test_low_level_traverse(self):
        branchy = load("branchy_glycan")
        t1 = monosaccharide.traverse(branchy.root)
        t2 = branchy.iternodes()
        for a, b in zip(list(t1), list(t2)):
            self.assertEqual(a, b)

    def test_low_level_graph_clone(self):
        branchy = load("branchy_glycan")
        self.assertEqual(branchy.root, monosaccharide.graph_clone(branchy.root))

    def test_ring_shape(self):
        hexose = named_structures.monosaccharides.Hex
        self.assertEqual(hexose.ring_type, "pyranose")
        hexose.ring_start += 1
        self.assertEqual(hexose.ring_type, "furanose")
        hexose.ring_start = 0
        hexose.ring_end = 0
        self.assertEqual(hexose.ring_type, "open")
        hexose.ring_end = None
        self.assertEqual(hexose.ring_type, "x")

    def test_exact_ordering_equality(self):
        hexose = named_structures.monosaccharides.Hex
        self.assertEqual(hexose, hexose)
        res1, res2 = hexose.clone(), hexose.clone()

        self.assertEqual(res1, res2)
        res1.add_substituent("n_acetyl", 2)
        res2.add_substituent("n_acetyl", 2)
        self.assertEqual(res1, res2)
        res2.drop_substituent(2)
        res2.add_substituent("n_acetyl", 3)
        self.assertNotEqual(res1, res2)

        res3 = res1.clone()
        res1.add_monosaccharide(hexose.clone(), 6)
        res3.add_monosaccharide(hexose.clone(), 6)
        self.assertEqual(res1, res3)

        res3.drop_monosaccharide(6)
        res3.add_monosaccharide(hexose.clone(), 5)
        self.assertNotEqual(res1, res3)

    def test_children(self):
        n_core = named_structures.glycans["N-Linked Core"]
        n_core.set_reducing_end(True)
        for pos, child in n_core.root.children():
            self.assertNotEqual(type(child), substituent.Substituent)
        for pos, child in n_core.root.reducing_end.children():
            self.assertNotEqual(type(child), Monosaccharide)

    def test_total_composition(self):
        hexose = named_structures.monosaccharides.Hex
        self.assertEqual(hexose.total_composition(), {"C": 6, "O": 6, "H": 12})
        hexose.reducing_end = True
        self.assertEqual(hexose.total_composition(), {"C": 6, "O": 6, "H": 14})

    def test_serialize_does_not_mutate(self):
        ref = self.glycan.clone()
        ids = [node.id for node in ref]
        for i, node in enumerate(ref):
            self.assertEqual(ids[i], node.id)
            self.assertEqual(node, self.glycan[i])
            str(node)
            self.assertEqual(ids[i], node.id)
            self.assertEqual(node, self.glycan[i])

    def test_pickle(self):
        for mono in named_structures.monosaccharides.values():
            self.assertEqual(mono, pickle.loads(pickle.dumps(mono)))


class ReducedEndTests(unittest.TestCase):
    def test_equality(self):
        self.assertEqual(ReducedEnd(), ReducedEnd())
        self.assertNotEqual(ReducedEnd(), ReducedEnd("H[2]H"))


if __name__ == '__main__':
    unittest.main()



import pkg_resources
import logging
from rdflib import Namespace, Graph, Literal
from rdflib.namespace import RDF, RDFS, FOAF

logging.getLogger("rdflib.term").addHandler(logging.NullHandler())
logging.getLogger("rdflib.term").propagate = False


class PopulatedNamespace(Namespace):
    def __new__(cls, value, members):
        inst = super(cls, cls).__new__(cls, value)
        inst._members = members
        return inst

    def __dir__(self):
        members = dir(super(PopulatedNamespace, self))
        members = sorted(set(list(members) + self._members))
        return members


def parse_owl(stream):
    from lxml import etree

    NSGlycan = None

    def iri(element):
        return NSGlycan[element.attrib['IRI'].strip("#")]

    def abbrev_iri(element):
        return NSGlycan[element.attrib['abbreviatedIRI'].strip("#")]

    tree = etree.parse(stream)
    root = tree.getroot()

    BASE_URI = "http://purl.jp/bio/12/glyco/glycan"

    entities = []
    for elt in root.iter("{http://www.w3.org/2002/07/owl#}Declaration"):
        decl = elt[0]
        if "IRI" in decl.attrib:
            entities.append(decl.attrib['IRI'].strip("#"))
    NSGlycan = PopulatedNamespace(BASE_URI + "#", entities)

    GlycoRDF = Graph()
    for elt in root.iter("{http://www.w3.org/2002/07/owl#}ClassAssertion"):
        cls = iri(elt.find("{http://www.w3.org/2002/07/owl#}Class"))
        inst = iri(elt.find("{http://www.w3.org/2002/07/owl#}NamedIndividual"))
        GlycoRDF.add((inst, RDF.type, cls))

    for elt in root.iter("{http://www.w3.org/2002/07/owl#}ObjectPropertyAssertion"):
        prop = iri(elt.find("{http://www.w3.org/2002/07/owl#}ObjectProperty"))
        subj, val = map(iri, elt.findall(
            "{http://www.w3.org/2002/07/owl#}NamedIndividual"))
        GlycoRDF.add((subj, prop, val))

    for elt in root.iter("{http://www.w3.org/2002/07/owl#}DataPropertyAssertion"):
        prop = iri(elt.find("{http://www.w3.org/2002/07/owl#}DataProperty"))
        subj = iri(elt.find("{http://www.w3.org/2002/07/owl#}NamedIndividual"))
        val = elt.find("{http://www.w3.org/2002/07/owl#}Literal").text
        GlycoRDF.add((subj, prop, Literal(val)))

    for elt in root.iter("{http://www.w3.org/2002/07/owl#}SubClassOf"):
        try:
            cls1, cls2 = map(iri, elt.findall(
                "{http://www.w3.org/2002/07/owl#}Class"))
        except (KeyError, ValueError):
            continue
        GlycoRDF.add((cls1, RDFS.subClassOf, cls2))

    for elt in root.iter("{http://www.w3.org/2002/07/owl#}AnnotationAssertion"):
        annotation = elt.find(
            "{http://www.w3.org/2002/07/owl#}AnnotationProperty").attrib['abbreviatedIRI']
        ns, tp = annotation.split(":")
        if ns == "rdf":
            rel = RDF[tp]
        elif ns == 'rdfs':
            rel = RDFS[tp]
        elif ns == 'foaf':
            rel = FOAF[tp]
        else:
            raise ValueError(annotation)
        try:
            subj = NSGlycan[elt.find("{http://www.w3.org/2002/07/owl#}IRI").text.strip("#")]
        except AttributeError:
            continue
        val = elt.find("{http://www.w3.org/2002/07/owl#}Literal").text
        GlycoRDF.add((subj, rel, Literal(val)))

    return NSGlycan, GlycoRDF


NSGlycan, glycordf = parse_owl(
    pkg_resources.resource_stream(__name__, "data/glycan.owl"))


__all__ = [
    "NSGlycan",
    "glycordf",
]





from glypy.structure.named_structures import monosaccharides, glycans, motifs, monosaccharide_residues
from glypy.structure import Glycan, Monosaccharide, Substituent, Link, ReducedEnd
from glypy.composition import Composition
from glypy.structure import glycan_composition
from glypy.structure.glycan_composition import GlycanComposition, MonosaccharideResidue
from glypy.utils import root, tree
from glypy.utils.multimap import OrderedMultiMap


__all__ = [
    "composition", "io", "structure", "utils", 'algorithms',
    "monosaccharides", "glycans", "motifs", "monosaccharide_residues",
    "Glycan", "Monosaccharide", "Substituent", "Link", "ReducedEnd",
    "Composition", "GlycanComposition", "MonosaccharideResidue",
    "root", "tree", "OrderedMultiMap", "glycan_composition"
]



import re
from collections import namedtuple

from six import string_types as basestring

from glypy.composition import Composition
from glypy.structure import substituent as _substituent
from glypy.io.tree_builder_utils import try_int

from .carbon_descriptors import CarbonDescriptors
from .substituent_conversion import alin_to_substituent, substituent_to_alin


_NodeTypeSpec = namedtuple("NodeTypeSpec", [
    "carbon_descriptor",
    "substituents"
])


class NodeTypeSpec(_NodeTypeSpec):
    @classmethod
    def parse(cls, text, version_number):
        if version_number < 2.0:
            raise TypeError("Cannot parse type version earlier than 2.0")
        parts = text.split("_")
        skeleton_anomer = parts[0]
        substituents = []
        if len(parts) > 1:
            ring_specification_or_substituents = parts[1]
            if re.match(r"[0-9\-1?]+-[0-9\-1?]+", ring_specification_or_substituents):
                ring_specification = ring_specification_or_substituents
            else:
                ring_specification = '?-?'
                substituents.append(ring_specification_or_substituents)
        else:
            ring_specification = '?-?'
        substituents.extend(parts[2:])
        try:
            skeleton, anomer = skeleton_anomer.split("-")
        except Exception:
            skeleton = skeleton_anomer
            anomer = (-1, None)
        ring_start, ring_end = map(try_int, ring_specification.split("-"))
        anomeric_position, anomer_type = anomer
        return cls(CarbonDescriptors(skeleton, anomer_type, anomeric_position, ring_start, ring_end),
                   cls.translate_substituents(substituents))

    @classmethod
    def from_monosaccharide(cls, monosaccharide):
        descriptors = CarbonDescriptors.from_monosaccharide(monosaccharide)
        substituents = []
        for position, substituent in monosaccharide.substituents():
            substituents.append((position, substituent.name))
        return cls(descriptors, substituents)

    @classmethod
    def translate_substituents(self, substituents):
        result = []
        for map_spec in substituents:
            position = map_spec[0]
            if position == "?":
                position = -1
                i = 1
            else:
                i = 0
                position = ''
                while map_spec[i].isdigit():
                    position += map_spec[i]
                    i += 1
                position = int(position)
            map_code = map_spec[1:]
            record = alin_to_substituent(map_code)
            result.append((position, record.name))
        return result

    def to_monosaccharide(self):
        base = self.carbon_descriptor.to_base_type()
        child_loss = Composition("H")
        for position, subst in self.substituents:
            try:
                parent_loss = _substituent.attachment_composition_info[subst]
            except KeyError:
                parent_loss = _substituent.default_attachment_composition
            base.add_substituent(subst, position, parent_loss=parent_loss, child_loss=child_loss)
        return base

    def to_res(self):
        mods = []
        for position, substituent in self.substituents:
            alin = substituent_to_alin(substituent)
            if position == -1:
                position = "?"
            mods.append("%s*%s" % (position, alin))
        encoded = '_'.join([self.carbon_descriptor.to_backbone_code()] + mods)
        return encoded

    def __str__(self):
        return self.to_res()

    def __repr__(self):
        return super(NodeTypeSpec, self).__repr__()

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, basestring):
            return str(self) == other
        if self.carbon_descriptor != other.carbon_descriptor:
            return False
        elif self.substituents != other.substituents:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.carbon_descriptor)



from collections import defaultdict
from weakref import ref

from six import string_types as basestring

from glypy.utils import uid
from glypy.composition import Composition

from . import atomic_data


class Graph(object):
    def __init__(self, vertices=None, edges=None):
        self.vertices = vertices or {}
        self.edges = edges or set()
        self.adjacency_list = defaultdict(set)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.vertices}, {self.edges})"
        return template.format(self=self)

    def add_vertex(self, vertex):
        self.vertices[vertex.id] = vertex
        vertex.graph = ref(self)
        return self

    def __getitem__(self, i):
        return self.vertices[i]

    def __iter__(self):
        keys = sorted(self.vertices.keys())
        for key in keys:
            yield self.vertices[key]

    def __len__(self):
        return len(self.vertices)

    def add_edge(self, edge):
        self.edges.add(edge)
        self.adjacency_list[edge.vertex_ids[0]].add((edge.vertex_ids[1], edge))
        self.adjacency_list[edge.vertex_ids[1]].add((edge.vertex_ids[0], edge))
        return self

    def neighbors_of(self, vertex_id):
        neighbors = [self[x[0]] for x in self.adjacency_list[vertex_id]]
        return neighbors

    def edges_of(self, vertex_id):
        edges = [x[1] for x in self.adjacency_list[vertex_id]]
        return edges


class Vertex(object):
    def __init__(self, id=None, _graph=None, **node_data):
        if id is None:
            id = uid()
        self.id = id
        self.data = node_data
        self._graph = None
        self.graph = _graph

    @property
    def graph(self):
        if self._graph is not None:
            return self._graph()
        return None

    @graph.setter
    def graph(self, value):
        self._graph = value

    def __repr__(self):
        if self.data:
            data = ', {self.data}'.format(self=self)
        else:
            data = ''
        template = "{self.__class__.__name__}({self.id}{data})"
        return template.format(self=self, data=data)

    def neighbors(self):
        graph = self.graph
        if graph is None:
            return []
        neighbors = graph.neighbors_of(self.id)
        return neighbors


class Edge(object):
    def __init__(self, vertex_ids):
        self.vertex_ids = tuple(vertex_ids)

    def __repr__(self):
        template = '{self.__class__.__name__}({self.vertex_ids})'
        return template.format(self=self)

    def __eq__(self, other):
        return self.vertex_ids == other.vertex_ids

    def __hash__(self):
        return hash(self.vertex_ids)


class Atom(Vertex):
    def __init__(self, element, id=None, _graph=None, **node_data):
        if isinstance(element, basestring):
            element = atomic_data.element_information[element]
        super(Atom, self).__init__(id, _graph=_graph, **node_data)
        self.element = element

    def __repr__(self):
        if self.data:
            data = ', {self.data}'.format(self=self)
        else:
            data = ''
        template = "{self.__class__.__name__}({self.element.symbol!r}, {self.id}{data})"
        return template.format(self=self, data=data)

    @property
    def valence(self):
        return self.element.valence

    @property
    def unpaired_electrons(self):
        theoretical = self.valence
        graph = self.graph
        bonds = graph.edges_of(self.id)
        for bond in bonds:
            theoretical -= bond.order
        return theoretical


class Bond(Edge):
    def __init__(self, vertex_ids, order=1):
        super(Bond, self).__init__(vertex_ids)
        self.order = order

    def __repr__(self):
        template = '{self.__class__.__name__}({self.vertex_ids}, {self.order})'
        return template.format(self=self)


class MolecularGraph(Graph):

    def total_composition(self):
        composition = Composition()
        unpaired_electrons = 0
        for node in self:
            composition[node.element.symbol] += 1
            unpaired_electrons += node.unpaired_electrons
        composition['H'] = unpaired_electrons
        return composition

    def mass(self):
        return self.total_composition().mass
